"""Custom Warning for GAPs"""

import logging

logger = logging.getLogger("gaps")


class gapsWarning(UserWarning):  # noqa: N801
    """Generic gaps Warning"""

    def __init__(self, *args, **kwargs):
        """Init exception and broadcast message to logger"""
        super().__init__(*args, **kwargs)
        if args:
            logger.warning(str(args[0]), stacklevel=2)


class gapsCollectionWarning(gapsWarning):  # noqa: N801
    """gaps Collection waring"""


class gapsHPCWarning(gapsWarning):  # noqa: N801
    """gaps HPC warning"""


class gapsDeprecationWarning(gapsWarning, DeprecationWarning):  # noqa: N801
    """gaps deprecation warning"""



"""gaps Job status manager"""

import json
import time
import shutil
import pprint
import logging
import datetime as dt
from pathlib import Path
from copy import deepcopy
from warnings import warn
from itertools import chain
from collections import UserDict, abc

import pandas as pd
import numpy as np

from rex.utilities import safe_json_load
from gaps.hpc import SLURM, PBS
from gaps.config import ConfigType
from gaps.utilities import recursively_update_dict, CaseInsensitiveEnum
from gaps.exceptions import gapsKeyError, gapsTypeError
from gaps.warn import gapsWarning


logger = logging.getLogger(__name__)
DT_FMT = "%d-%b-%Y %H:%M:%S"


class StatusField(CaseInsensitiveEnum):
    """A collection of required status fields in a status file"""

    JOB_ID = "job_id"
    JOB_STATUS = "job_status"
    PIPELINE_INDEX = "pipeline_index"
    HARDWARE = "hardware"
    QOS = "qos"
    OUT_FILE = "out_file"
    TIME_SUBMITTED = "time_submitted"
    TIME_START = "time_start"
    TIME_END = "time_end"
    TOTAL_RUNTIME = "total_runtime"
    RUNTIME_SECONDS = "runtime_seconds"
    MONITOR_PID = "monitor_pid"
    STDOUT_LOG = "stdout_log"
    STDOUT_ERR_LOG = "stdout_err_log"


class HardwareOption(CaseInsensitiveEnum):
    """A collection of hardware options"""

    LOCAL = "local"
    """Local execution"""
    KESTREL = "kestrel"
    """NREL's Kestrel HPC. Assumes SLURM scheduler"""
    EAGLE = "eagle"
    """NREL's Eagle HPC. Assumes SLURM scheduler"""
    AWSPC = "awspc"
    """AWS Parallel Cluster. Assumes SLURM scheduler"""
    SLURM = "slurm"
    """Fallback value for any HPC system that runs SLURM"""
    PEREGRINE = "peregrine"
    """NREL's Peregrine HPC. Assumes PBS scheduler"""

    @classmethod
    def _new_post_hook(cls, obj, value):
        """Hook for post-processing after __new__"""

        if value in {"eagle", "kestrel", "awspc", "slurm"}:
            obj.manager = SLURM()
        elif value == "peregrine":
            obj.manager = PBS()
        else:
            obj.manager = None

        return obj

    @property
    def is_hpc(self):
        """bool: Hardware option is HPC"""
        return self.value != "local"

    @property
    def charge_factor(self):
        """int: Hardware AU charge factor (per node-hour)"""
        if self.value == self.KESTREL:
            return 10
        if self.value == self.EAGLE:
            return 3
        if self.value == self.PEREGRINE:
            return 1
        return 0

    def check_status_using_job_id(self, *args, **kwargs):
        """Check the status of a job using a job ID.

        Parameters
        ----------
        args, kwargs
            Args and keyword-args to pass to
            `manager.check_status_using_job_id`.

        Returns
        -------
        status : str | None
            Queue job status string or ``None`` if not found.
        """
        if self.manager is None:
            return None
        return self.manager.check_status_using_job_id(*args, **kwargs)

    @property
    def supports_categorical_qos(self):
        """bool: Hardware option supports categorical QOS values"""
        return self.value in {self.EAGLE, self.KESTREL, self.AWSPC, self.SLURM}

    @classmethod
    def reset_all_cached_queries(cls):
        """Reset all cached hardware queries"""
        cls.EAGLE.manager.reset_query_cache()
        cls.KESTREL.manager.reset_query_cache()
        cls.AWSPC.manager.reset_query_cache()
        cls.SLURM.manager.reset_query_cache()
        cls.PEREGRINE.manager.reset_query_cache()


class StatusOption(CaseInsensitiveEnum):
    """A collection of job status options"""

    NOT_SUBMITTED = "not submitted"
    SUBMITTED = "submitted"
    RUNNING = "running"
    SUCCESSFUL = "successful"
    FAILED = "failed"
    COMPLETE = "complete"

    @classmethod
    def _new_post_hook(cls, obj, value):
        """Hook for post-processing after __new__"""
        verb = "has" if value == "failed" else "is"
        obj.with_verb = f"{verb} {value}"
        obj.is_processing = value in {"submitted", "running"}
        return obj


class QOSOption(CaseInsensitiveEnum):
    """A collection of job QOS options"""

    NORMAL = "normal"
    """Normal QOS"""
    HIGH = "high"
    """High QOS"""
    STANDBY = "standby"
    """Standby QOS"""
    UNSPECIFIED = "unspecified"
    """Unspecified QOS"""

    @classmethod
    def _new_post_hook(cls, obj, value):
        """Hook for post-processing after __new__"""
        obj.charge_factor = 2 if value == "high" else 1
        return obj


class HardwareStatusRetriever:
    """Query hardware for job status"""

    def __init__(self, subprocess_manager=None):
        """Initialize `HardwareStatusRetriever`.

        Parameters
        ----------
        subprocess_manager : PBS | SLURM | None, optional
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC. By default, `None`.
        """
        self.subprocess_manager = subprocess_manager

    def __getitem__(self, key):
        job_id, hardware = key
        if not job_id:
            return None

        if self.subprocess_manager is not None:
            return self.subprocess_manager.check_status_using_job_id(job_id)

        return _validate_hardware(hardware).check_status_using_job_id(job_id)


class Status(UserDict):
    """Base class for data pipeline health and status information.

    This class facilitates the creating of individual job status files
    which the main `Status` object can collect.
    """

    _DF_COLUMNS = [
        StatusField.JOB_STATUS.value,
        StatusField.PIPELINE_INDEX.value,
        StatusField.JOB_ID.value,
        StatusField.TIME_SUBMITTED.value,
        StatusField.TIME_START.value,
        StatusField.TIME_END.value,
        StatusField.TOTAL_RUNTIME.value,
        StatusField.HARDWARE.value,
        StatusField.QOS.value,
    ]
    HIDDEN_SUB_DIR = ".gaps"
    MONITOR_PID_FILE = "monitor_pid.json"
    JOB_STATUS_FILE = "jobstatus_{}.json"
    NAMED_STATUS_FILE = "{}_status.json"

    def __init__(self, status_dir):
        """Initialize `Status`.

        Parameters
        ----------
        status_dir : path-like
            Directory containing zero or more job json status files.
        """
        super().__init__()
        self.dir = Path(status_dir).expanduser().resolve()
        self._validate_dir()
        self.name = self.dir.name
        self.dir /= self.HIDDEN_SUB_DIR
        self._fpath = self.dir / self.NAMED_STATUS_FILE.format(self.name)
        self.data = _load(self._fpath)

    def _validate_dir(self):
        """Validate that the directory name is not a config file type"""
        for file_type in ConfigType.members_as_str():
            if self.dir.name.endswith(f".{file_type}"):
                msg = (
                    f"Need a directory containing a status {file_type}, "
                    f"not a status {file_type}: {self.dir!r}"
                )
                raise gapsTypeError(msg)

    @property
    def job_ids(self):
        """list: Flat list of job ids"""
        return _get_attr_flat_list(self.data, key=StatusField.JOB_ID)

    @property
    def job_hardware(self):
        """list: Flat list of job hardware options"""
        return _get_attr_flat_list(self.data, key=StatusField.HARDWARE)

    def as_df(self, pipe_steps=None, index_name="job_name", include_cols=None):
        """Format status as pandas DataFrame.

        Parameters
        ----------
        pipe_steps : container, optional
            A container of pipeline step names to collect. If `None`,
            all pipeline steps in the status file are collected.
            By default, `None`.
        index_name : str, optional
            Name to assign to index of DataFrame.
            By default, `"job_name"`.

        Returns
        -------
        pd.DataFrame
            Pandas DataFrame containing status information.
        """
        include_cols = include_cols or []
        output_cols = self._DF_COLUMNS + list(include_cols)

        self.update_from_all_job_files(purge=False, check_hardware=True)
        if not self.data:
            return pd.DataFrame(columns=output_cols)

        data = deepcopy(self.data)
        requested_steps = pipe_steps or self.keys()
        steps = []
        for step, status in data.items():
            if step not in requested_steps:
                continue
            try:
                step_index = status.pop(StatusField.PIPELINE_INDEX, None)
            except (AttributeError, TypeError):
                continue
            if not status:
                status = {step: {}}  # noqa: PLW2901
            try:
                step_df = pd.DataFrame(status).T
            except ValueError:
                continue
            step_df[f"{StatusField.PIPELINE_INDEX}"] = step_index
            steps.append(step_df)

        try:
            step_df = pd.concat(steps, sort=False)
        except ValueError:
            return pd.DataFrame(columns=output_cols)

        for field in chain(StatusField, include_cols):
            if field not in step_df.columns:
                step_df[f"{field}"] = np.nan

        step_df.loc[
            step_df[StatusField.JOB_STATUS].isna(),
            StatusField.JOB_STATUS.value,
        ] = StatusOption.NOT_SUBMITTED.value

        step_df = _add_elapsed_time(step_df)

        step_df.index.name = index_name
        return step_df[output_cols]

    def reload(self):
        """Re-load the data from disk"""
        self.data = _load(self._fpath)

    def reset_after(self, pipeline_step):
        """Reset status of all pipeline steps after the input one.

        Parameters
        ----------
        pipeline_step : str
            Pipeline step to delineate which parts of the status should
            be reset. If this pipeline step is not found in the status,
            nothing is reset. The status for the pipeline step is
            untouched; only the status of steps following this one are
            reset.
        """
        reset_index = self.step_index(pipeline_step)
        if reset_index is None:
            return

        for step_name, step_status in self.items():
            try:
                step_index = step_status.get(StatusField.PIPELINE_INDEX)
            except AttributeError:
                continue

            if step_index is None:
                continue

            if step_index > reset_index:
                self.data[step_name] = {StatusField.PIPELINE_INDEX: step_index}

    def step_index(self, pipeline_step):
        """Get pipeline index for the pipeline step, if it exists.

        Parameters
        ----------
        pipeline_step : str
            Name of pipeline step.

        Returns
        -------
        int | None
            Pipeline index of pipeline step if it is found in the
            status, ``None`` otherwise.
        """
        step_status = self.data.get(pipeline_step, {})
        return step_status.get(StatusField.PIPELINE_INDEX)

    def dump(self):
        """Dump status json w/ backup file in case process is killed"""

        self._fpath.parent.mkdir(parents=True, exist_ok=True)

        backup = self._fpath.name.replace(".json", "_backup.json")
        backup = self._fpath.parent / backup
        if self._fpath.exists():
            shutil.copyfile(self._fpath, backup)

        with self._fpath.open("w", encoding="utf-8") as status:
            json.dump(self.data, status, indent=4, separators=(",", ": "))

        backup.unlink(missing_ok=True)

    def update_from_all_job_files(self, check_hardware=False, purge=True):
        """Update status from all single-job job status files

        This method loads all single-job status files in the target
        directory and updates the `Status` object with the single-job
        statuses.

        Parameters
        ----------
        check_hardware : bool, optional
            Option to check hardware status for job failures for jobs
            with a "running" status. This is useful because the
            "running" status may not be correctly updated if the job
            terminates abnormally on the HPC. By default, `False`.
        purge : bool, optional
            Option to purge the individual status files.
            By default, `True`.

        Returns
        -------
        `Status`
            This instance of `Status` with updated job properties.
        """
        file_pattern = self.JOB_STATUS_FILE.format("*")
        for file_ in Path(self.dir).glob(file_pattern):
            status = _safe_load(file_, purge=purge)
            self.data = recursively_update_dict(self.data, status)

        monitor_pid_file = Path(self.dir) / self.MONITOR_PID_FILE
        if monitor_pid_file.exists():
            monitor_pid_info = _safe_load(monitor_pid_file, purge=purge)
            self.data.update(monitor_pid_info)

        if check_hardware:
            self._update_from_hardware()

        if purge:
            self.dump()

        return self

    def _update_from_hardware(self):
        """Check all job status against hardware status"""
        hardware_status_retriever = HardwareStatusRetriever()
        for job_data in self._job_statuses():
            self._update_job_status_from_hardware(
                job_data, hardware_status_retriever
            )

    def _job_statuses(self):
        """Iterate over job statuses. Ignore other info in self.data"""
        for status in self.values():
            try:
                yield from _iter_job_status(status)
            except AttributeError:  # noqa: PERF203
                continue

    @staticmethod
    def _update_job_status_from_hardware(job_data, hardware_status_retriever):
        """Update job status to failed if processing but DNE on HPC"""

        status = job_data.get(
            StatusField.JOB_STATUS, StatusOption.NOT_SUBMITTED
        )
        try:
            if not StatusOption(status).is_processing:
                return
        except ValueError:
            pass

        job_id = job_data.get(StatusField.JOB_ID, None)
        job_hardware = job_data.get(StatusField.HARDWARE, None)

        # get job status from hardware
        current = hardware_status_retriever[job_id, job_hardware]
        # No current status and job was not successful: failed!
        if current is None:
            job_data[StatusField.JOB_STATUS] = StatusOption.FAILED

    def update_job_status(
        self, pipeline_step, job_name, hardware_status_retriever=None
    ):
        """Update single-job job status from single-job job status file.

        If the status for a given pipeline step/job name combination is
        not found, the status object remains unchanged.

        Parameters
        ----------
        pipeline_step : str
            Pipeline step that the job belongs to.
        job_name : str
            Unique job name identification.
        hardware_status_retriever : `HardwareStatusRetriever`, optional
            Hardware status retriever. By default, `None`, which creates
            an instance internally.
        """
        hardware_status_retriever = (
            hardware_status_retriever or HardwareStatusRetriever()
        )
        # look for job completion file.
        current = _load_job_file(self.dir, job_name)

        # Update status data dict and file if job file was found
        if current is not None:
            self.data = recursively_update_dict(self.data, current)

        # check job status via hardware if job file not found.
        elif pipeline_step in self.data:
            # job exists
            if job_name in self.data[pipeline_step]:
                self._update_job_status_from_hardware(
                    self.data[pipeline_step][job_name],
                    hardware_status_retriever,
                )
            # job does not yet exist
            else:
                self.data[pipeline_step][job_name] = {
                    StatusField.JOB_STATUS: StatusOption.NOT_SUBMITTED
                }

        self.dump()

    def _retrieve_job_status(
        self, pipeline_step, job_name, hardware_status_retriever
    ):
        """Update and retrieve job status"""
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        self.update_job_status(
            pipeline_step, job_name, hardware_status_retriever
        )

        try:
            job_data = self[pipeline_step][job_name]
        except KeyError:
            return None

        return job_data.get(StatusField.JOB_STATUS)

    @classmethod
    def _dump_dict(cls, status_dir, file_name, out_dict):
        """Dump the dict to a file, making sure dirs exist"""
        fpath = Path(status_dir) / cls.HIDDEN_SUB_DIR / file_name
        fpath.parent.mkdir(parents=True, exist_ok=True)
        with fpath.open("w", encoding="utf-8") as out_file:
            json.dump(
                out_dict,
                out_file,
                sort_keys=True,
                indent=4,
                separators=(",", ": "),
            )

    @classmethod
    def record_monitor_pid(cls, status_dir, pid):
        """Make a json file recording the PID of the monitor process

        Parameters
        ----------
        status_dir : path-like
            Directory to put json status file.
        pid : int
            PID of the monitoring process.
        """
        pid = {StatusField.MONITOR_PID: pid}
        cls._dump_dict(status_dir, cls.MONITOR_PID_FILE, pid)

    @classmethod
    def make_single_job_file(cls, status_dir, pipeline_step, job_name, attrs):
        """Make a json file recording the status of a single job

        This method should primarily be used by HPC nodes to mark the
        status of individual jobs.

        Parameters
        ----------
        status_dir : path-like
            Directory to put json status file.
        pipeline_step : str
            Pipeline step that the job belongs to.
        job_name : str
            Unique job name identification.
        attrs : dict
            Dictionary of job attributes that represent the job status
            attributes.
        """
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        status = {pipeline_step: {job_name: attrs}}
        out_fn = cls.JOB_STATUS_FILE.format(job_name)
        cls._dump_dict(status_dir, out_fn, status)

    @classmethod
    def mark_job_as_submitted(
        cls, status_dir, pipeline_step, job_name, replace=False, job_attrs=None
    ):
        """Mark a job in the status json as "submitted"

        Status json does not have to exist - it is created if missing.
        If it exists, the job will only be updated if it does not
        exist (i.e. not submitted), unless replace is set to `True`.

        Parameters
        ----------
        status_dir : path-like
            Directory containing json status file.
        pipeline_step : str
            Pipeline step that the job belongs to.
        job_name : str
            Unique job name identification.
        replace : bool, optional
            Flag to force replacement of pre-existing job status.
            By default, `False`.
        job_attrs : dict, optional
            Job attributes. Should include 'job_id' if running on HPC.
            By default, `None`.
        """
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        obj = cls(status_dir)

        job_attrs = job_attrs or {}
        hardware = job_attrs.get(StatusField.HARDWARE)
        try:
            job_on_hpc = HardwareOption(hardware).is_hpc
        except ValueError:
            job_on_hpc = False

        if job_on_hpc and StatusField.JOB_ID not in job_attrs:
            msg = (
                f'Key "job_id" should be in kwargs for {job_name!r} if '
                f"adding job from an HPC node."
            )
            warn(msg, gapsWarning)

        exists = obj.job_exists(
            status_dir, job_name, pipeline_step=pipeline_step
        )

        if exists and not replace:
            return

        job_status = job_attrs.get(StatusField.JOB_STATUS)
        if job_status != StatusOption.SUBMITTED:
            if job_status is not None:
                msg = (
                    f"Attempting to mark a job as submitted but included a "
                    f"{StatusField.JOB_STATUS} value of {job_status!r} in the "
                    f"job_attrs dictionary! Setting the job status to "
                    f"{StatusOption.SUBMITTED!r} before writing."
                )
                warn(msg, gapsWarning)
            job_attrs[StatusField.JOB_STATUS] = StatusOption.SUBMITTED

        if pipeline_step not in obj.data:
            obj.data[pipeline_step] = {job_name: job_attrs}
        else:
            obj.data[pipeline_step][job_name] = job_attrs

        obj.dump()

    @classmethod
    def job_exists(cls, status_dir, job_name, pipeline_step=None):
        """Check whether a job exists and return a bool

        This method will return `True` if the job name is found as a
        key in the dictionary under the `pipeline_step` keyword, or any
        pipeline step if a `None` value is passed.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        job_name : str
            Unique job name identification.
        pipeline_step : str, optional
            Pipeline step that the job belongs to. By default, `None`,
            which checks all pipeline steps for the job name.

        Returns
        -------
        exists : bool
            `True` if the job exists in the status json.
        """
        if job_name.endswith(".h5"):
            job_name = job_name.replace(".h5", "")

        obj = cls(status_dir).update_from_all_job_files(purge=False)
        if not obj.data:
            return False

        if pipeline_step is not None:
            jobs = [obj.data.get(pipeline_step)]
        else:
            jobs = obj.data.values()

        for job in jobs:
            if not job:
                continue
            for name in job:
                if name == job_name:
                    return True

        return False

    @classmethod
    def retrieve_job_status(
        cls, status_dir, pipeline_step, job_name, subprocess_manager=None
    ):
        """Update and retrieve job status

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        pipeline_step : str
            Pipeline step that the job belongs to.
        job_name : str
            Unique job name identification.
        subprocess_manager : None | SLURM, optional
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC.

        Returns
        -------
        status : str | None
            Status string or `None` if job/pipeline step not found.
        """
        hsr = HardwareStatusRetriever(subprocess_manager)
        return cls(status_dir)._retrieve_job_status(  # noqa: SLF001
            pipeline_step, job_name, hsr
        )

    @classmethod
    def parse_step_status(
        cls, status_dir, pipeline_step, key=StatusField.OUT_FILE
    ):
        """Parse key from job status(es) from the given pipeline step

        This command DOES NOT check the HPC queue for jobs and therefore
        DOES NOT update the status of previously running jobs that have
        errored out of the HPC queue.

        Parameters
        ----------
        status_dir : path-like
            Directory containing the status file to parse.
        pipeline_step : str
            Target pipeline step to parse.
        key : StatusField | str, optional
            Parsing target of previous pipeline step. By default,
            `StatusField.OUT_FILE`.

        Returns
        -------
        list
            Arguments parsed from the status file in status_dir from
            the input pipeline step. This list is empty if the `key` is
            not found in the job status, or if the pipeline step does
            not exist in status.
        """
        status = cls(status_dir).update_from_all_job_files(purge=False)
        step_status = status.get(pipeline_step, {})
        step_status.pop(StatusField.PIPELINE_INDEX, None)
        return _get_attr_flat_list(step_status, key=key)


class StatusUpdates:
    """Context manager to track run function progress.

    When this context is entered, a status file is written for the given
    pipeline step/job combination, with the given job attributes. The
    job status is set to "running", and the start time is recorded. When
    the context is exited, another status file is written with the end
    time and total runtime values added. The status is also set to
    "successful", unless an uncaught exception was raised during the
    function runtime, in which case the status is set to "failed". If
    the `out_file` attribute of this context manager is set before the
    context is exited, that value is also written to the status file.
    """

    def __init__(self, directory, pipeline_step, job_name, job_attrs):
        """Initialize `StatusUpdates`.

        Parameters
        ----------
        directory : path-like
            Directory to write status files to.
        pipeline_step : str
            Name of the pipeline step being run.
        job_name : str
            Name of the job being run.
        job_attrs : dict
            A dictionary containing job attributes that should be
            written to the status file.
        """
        self.directory = directory
        self.pipeline_step = pipeline_step
        self.job_name = job_name
        self.job_attrs = deepcopy(job_attrs)
        self.start_time = None
        self.out_file = None

    def __enter__(self):
        logger.debug(
            "Received job attributes: %s",
            pprint.pformat(self.job_attrs, indent=4),
        )

        self.start_time = dt.datetime.now()
        self.job_attrs.update(
            {
                StatusField.JOB_STATUS: StatusOption.RUNNING,
                StatusField.TIME_START: self.start_time.strftime(DT_FMT),
            }
        )
        Status.make_single_job_file(
            self.directory,
            pipeline_step=self.pipeline_step,
            job_name=self.job_name,
            attrs=self.job_attrs,
        )
        return self

    def __exit__(self, exc_type, exc, traceback):
        end_time = dt.datetime.now()
        time_elapsed_s = (end_time - self.start_time).total_seconds()
        time_elapsed = _elapsed_time_as_str(time_elapsed_s)
        self.job_attrs.update(
            {
                StatusField.TIME_END: end_time.strftime(DT_FMT),
                StatusField.TOTAL_RUNTIME: time_elapsed,
                StatusField.RUNTIME_SECONDS: time_elapsed_s,
            }
        )
        if exc is None:
            self.job_attrs.update(
                {
                    StatusField.OUT_FILE: self.out_file,
                    StatusField.JOB_STATUS: StatusOption.SUCCESSFUL,
                }
            )
            logger.info(
                "Pipeline step %r complete. Time elapsed: %s. "
                "Target output file: %r",
                self.pipeline_step,
                time_elapsed,
                self.out_file,
            )
        else:
            self.job_attrs.update(
                {StatusField.JOB_STATUS: StatusOption.FAILED}
            )
            logger.info(
                "Pipeline step %r failed in %s",
                self.pipeline_step,
                time_elapsed,
            )

        Status.make_single_job_file(
            self.directory,
            pipeline_step=self.pipeline_step,
            job_name=self.job_name,
            attrs=self.job_attrs,
        )


def _elapsed_time_as_str(seconds_elapsed):
    """Format elapsed time into human readable string"""
    days, seconds = divmod(int(seconds_elapsed), 24 * 3600)
    minutes, seconds = divmod(seconds, 60)
    hours, minutes = divmod(minutes, 60)
    time_str = f"{hours:d}:{minutes:02d}:{seconds:02d}"
    if days:
        time_str = f"{days:,d} day{'s' if abs(days) != 1 else ''}, {time_str}"
    return time_str


def _add_elapsed_time(status_df):
    """Add elapsed time to status DataFrame"""
    has_start_time = ~status_df[StatusField.TIME_START].isna()
    has_no_end_time = status_df[StatusField.TIME_END].isna()
    has_not_failed = status_df[StatusField.JOB_STATUS] != StatusOption.FAILED
    mask = has_start_time & (has_no_end_time & has_not_failed)

    status_df = _add_time_cols_if_needed(status_df)
    start_times = status_df.loc[mask, StatusField.TIME_START]
    start_times = pd.to_datetime(start_times, format=DT_FMT)
    elapsed_times = dt.datetime.now() - start_times
    elapsed_times = elapsed_times.apply(lambda dt: dt.total_seconds())
    status_df.loc[mask, StatusField.RUNTIME_SECONDS] = elapsed_times
    elapsed_times = elapsed_times.apply(_elapsed_time_as_str)
    elapsed_times = elapsed_times.apply(lambda time_str: f"{time_str} (r)")
    status_df.loc[mask, StatusField.TOTAL_RUNTIME] = elapsed_times
    return status_df


def _add_time_cols_if_needed(status_df):
    """Adds any missing time cols to avoid pandas 2.0 warnings"""
    for col in [StatusField.RUNTIME_SECONDS, StatusField.TOTAL_RUNTIME]:
        if col not in status_df:
            status_df[col] = None
    return status_df


def _load(fpath):
    """Load status json"""
    if fpath.is_file():
        return safe_json_load(fpath.as_posix())
    return {}


def _load_job_file(status_dir, job_name, purge=True):
    """Load a single-job job status file in the target status_dir"""
    status_dir = Path(status_dir)
    status_fname = status_dir / Status.JOB_STATUS_FILE.format(job_name)
    if status_fname not in status_dir.glob("*"):
        return None
    return _safe_load(status_fname, purge=purge)


def _safe_load(file_path, purge=True):
    """Safe load json file and purge if needed"""
    # wait one second to make sure file is finished being written
    time.sleep(0.01)
    status = safe_json_load(file_path.as_posix())
    if purge:
        file_path.unlink()
    return status


def _get_attr_flat_list(inp, key=StatusField.JOB_ID):
    """Get all job attribute values from the status data dict"""

    out = []
    if not isinstance(inp, abc.Mapping):
        return out

    if key in inp:
        out = _combine_iterables_and_numerics(out, inp[key])
    else:
        for val in inp.values():
            temp = _get_attr_flat_list(val, key=key)
            out = _combine_iterables_and_numerics(out, temp)

    return out


def _combine_iterables_and_numerics(iterable, new_vals):
    """Combine new_vals into the "iterable" list"""
    try:
        return iterable + new_vals
    except TypeError:
        pass

    return [*iterable, new_vals]


def _validate_hardware(hardware):
    """Verify that the selected hardware is a valid option"""
    try:
        return HardwareOption(hardware)
    except ValueError as err:
        msg = (
            f"Requested hardware ({hardware!r}) not recognized! "
            f"Available options are: {HardwareOption.members_as_str()}."
        )
        raise gapsKeyError(msg) from err


def _iter_job_status(status):
    """Iterate over job status dictionary"""
    for job_status in status.values():
        if not isinstance(job_status, dict):
            continue
        yield job_status



"""GAPs pipeline CLI entry points"""

import os
import sys
import logging
from pathlib import Path
from warnings import warn

import click
import psutil

from gaps import Pipeline
from gaps.config import ConfigType, init_logging_from_config_file
from gaps.cli.documentation import _pipeline_command_help
from gaps.cli.command import _WrappedCommand
from gaps.status import Status, StatusField
from gaps.exceptions import gapsExecutionError
from gaps.warn import gapsWarning


logger = logging.getLogger(__name__)
FORKED_PID = 0


def _can_run_background():
    """Determine if GAPs can execute the pipeline in the background"""
    return hasattr(os, "setsid") and hasattr(os, "fork")


def template_pipeline_config(commands):
    """Generate a template pipeline config based on the commands"""
    pipeline = []
    for command in commands:
        command_name_config = command.name.replace("-", "_")
        sample_config_filename = (
            f"./config_{command_name_config}.{ConfigType.JSON}"
        )
        pipeline.append({command.name: sample_config_filename})
    return {
        "pipeline": pipeline,
        "logging": {"log_file": None, "log_level": "INFO"},
    }


@click.pass_context
def pipeline(
    ctx, config_file, cancel, monitor, background=False, recursive=False
):
    """Execute multiple steps in an analysis pipeline"""

    if recursive:
        _submit_recursive_pipelines(ctx, cancel, monitor, background)
        return

    if config_file is None:
        config_files = _find_pipeline_config_files(Path())
        if len(config_files) != 1:
            msg = (
                f"Could not determine config file - multiple (or no) files "
                f" detected with 'pipeline' in the name exist: {config_file}"
            )
            raise gapsExecutionError(msg)

        config_file = config_files[0]

    init_logging_from_config_file(config_file, background=background)
    _run_pipeline(ctx, config_file, cancel, monitor, background)


def _submit_recursive_pipelines(ctx, cancel, monitor, background):
    """Submit pipelines in all recursive subdirectories"""
    start_dir = Path()
    for sub_dir in start_dir.glob("**/"):
        config_files = _find_pipeline_config_files(sub_dir)
        if sub_dir.name == Status.HIDDEN_SUB_DIR:
            continue

        if len(config_files) > 1:
            msg = (
                f"Could not determine config file - multiple files detected "
                f"with 'pipeline' in the name in the {str(sub_dir)!r} "
                "directory!"
            )
            warn(msg, gapsWarning)
            continue
        if len(config_files) == 0:
            continue

        init_logging_from_config_file(config_files[0], background=background)
        _run_pipeline(ctx, config_files[0], cancel, monitor, background)


def _find_pipeline_config_files(directory):
    """Find all files matching *pipeline* in directory"""
    return [fp for fp in Path(directory).glob("*pipeline*") if fp.is_file()]


def _run_pipeline(ctx, config_file, cancel, monitor, background):
    """Run a GAPs pipeline for an existing config file"""

    if cancel:
        Pipeline.cancel_all(config_file)
        return

    if background:
        if not _can_run_background():
            msg = (
                "Cannot run pipeline in background on system that does not "
                "implement os.fork and os.setsid"
            )
            raise gapsExecutionError(msg)
        ctx.obj["LOG_STREAM"] = False
        pid = _kickoff_background(config_file)
        if pid == FORKED_PID:
            sys.exit()
        return

    project_dir = str(Path(config_file).parent.expanduser().resolve())
    status = Status(project_dir).update_from_all_job_files(purge=False)
    monitor_pid = status.get(StatusField.MONITOR_PID)
    if monitor_pid is not None and psutil.pid_exists(monitor_pid):
        msg = (
            f"Another pipeline in {project_dir!r} is running on monitor PID: "
            f"{monitor_pid}. Not starting a new pipeline execution."
        )
        warn(msg, gapsWarning)
        return

    if monitor:
        Status.record_monitor_pid(Path(config_file).parent, os.getpid())

    Pipeline.run(config_file, monitor=monitor)


def _kickoff_background(config_file):  # pragma: no cover
    """Kickoff a child process that runs pipeline in the background"""
    pid = os.fork()
    if pid == FORKED_PID:
        os.setsid()  # This creates a new session
        Pipeline.run(config_file, monitor=True)
    else:
        Status.record_monitor_pid(Path(config_file).parent, pid)
        click.echo(
            f"Kicking off pipeline job in the background. Monitor PID: {pid}"
        )
    return pid


def pipeline_command(template_config):
    """Generate a pipeline command"""
    params = [
        click.Option(
            param_decls=["--config_file", "-c"],
            default=None,
            help=_pipeline_command_help(template_config),
        ),
        click.Option(
            param_decls=["--cancel"],
            is_flag=True,
            help="Flag to cancel all jobs associated with a given pipeline.",
        ),
        click.Option(
            param_decls=["--monitor"],
            is_flag=True,
            help="Flag to monitor pipeline jobs continuously. Default is not "
            "to monitor (kick off jobs and exit).",
        ),
        click.Option(
            param_decls=["--recursive", "-r"],
            is_flag=True,
            help="Flag to recursively submit pipelines, starting from the "
            "current directory and checking every sub-directory therein. The "
            "`-c` option will be *completely ignored* if you use this option. "
            "Instead, the code will check every sub-directory for exactly one "
            "file with the word `pipeline` in it. If found, that file is "
            "assumed to be the pipeline config and is used to kick off the "
            "pipeline. In any other case, the directory is skipped.",
        ),
    ]
    if _can_run_background():
        params += [
            click.Option(
                param_decls=["--background"],
                is_flag=True,
                help="Flag to monitor pipeline jobs continuously in the "
                "background. Note that the stdout/stderr will not be "
                "captured, but you can set a pipeline 'log_file' to "
                "capture logs.",
            ),
        ]

    return _WrappedCommand(
        "pipeline",
        context_settings=None,
        callback=pipeline,
        params=params,
        help=(
            "Execute multiple steps in an analysis pipeline.\n\n"
            "The general structure for calling this CLI command is given "
            "below (add ``--help`` to print help info to the terminal)."
        ),
        epilog=None,
        short_help=None,
        options_metavar="[OPTIONS]",
        add_help_option=True,
        no_args_is_help=False,
        hidden=False,
        deprecated=False,
    )



"""Main CLI entry points"""

from functools import partial

import click

from gaps import Pipeline
from gaps.cli.batch import batch_command
from gaps.cli.templates import template_command
from gaps.cli.reset import reset_command
from gaps.cli.pipeline import pipeline_command, template_pipeline_config
from gaps.cli.collect import collect
from gaps.cli.script import script
from gaps.cli.config import from_config
from gaps.cli.command import (
    CLICommandFromFunction,
    _WrappedCommand,
)
from gaps.cli.documentation import _main_command_help
from gaps.cli.preprocessing import (
    preprocess_collect_config,
    preprocess_script_config,
)
from gaps.cli.status import status_command


class _CLICommandGenerator:
    """Generate commands from a list of configurations"""

    def __init__(self, command_configs):
        """

        Parameters
        ----------
        command_configs : list of :class:`CLICommandFromFunction`
            List of command configs to convert to click commands.
        """
        self.command_configs = command_configs
        self.commands, self.template_configs = [], {}

    def add_collect_command_configs(self):
        """Add collect command if the function is split spatially"""
        all_commands = []
        for command_configuration in self.command_configs:
            all_commands.append(command_configuration)
            if command_configuration.add_collect:
                collect_configuration = CLICommandFromFunction(
                    name=f"collect-{command_configuration.name}",
                    function=collect,
                    split_keys=[("_out_path", "_pattern")],
                    config_preprocessor=preprocess_collect_config,
                )
                all_commands.append(collect_configuration)
        self.command_configs = all_commands
        return self

    def add_script_command(self):
        """Add script command as an option"""
        script_configuration = CLICommandFromFunction(
            name="script",
            function=script,
            split_keys=["_cmd"],
            config_preprocessor=preprocess_script_config,
        )
        self.command_configs.append(script_configuration)
        return self

    def convert_to_commands(self):
        """Convert all of the command configs into click commands"""
        for command_config in self.command_configs:
            command = as_click_command(command_config)
            self.commands.append(command)
            Pipeline.COMMANDS[command_config.name] = command
            template_config = command_config.documentation.template_config
            self.template_configs[command_config.name] = template_config
        return self

    def add_pipeline_command(self):
        """Add pipeline command, which includes the previous commands"""
        tpc = template_pipeline_config(self.command_configs)
        pipeline = pipeline_command(tpc)
        self.commands = [pipeline, *self.commands]
        self.template_configs["pipeline"] = tpc
        return self

    def add_batch_command(self):
        """Add the batch command"""
        self.commands.append(batch_command())
        return self

    def add_status_command(self):
        """Add the status command"""
        self.commands.append(status_command())
        return self

    def add_template_command(self):
        """Add the config template command"""
        self.commands.append(template_command(self.template_configs))
        return self

    def add_reset_command(self):
        """Add the status reset command"""
        self.commands.append(reset_command())
        return self

    def generate(self):
        """Generate a list of click commands from input configs"""
        return (
            self.add_collect_command_configs()
            .add_script_command()
            .convert_to_commands()
            .add_pipeline_command()
            .add_batch_command()
            .add_status_command()
            .add_template_command()
            .add_reset_command()
            .commands
        )


def as_click_command(command_config):
    """Convert a GAPs command configuration into a ``click`` command

    Parameters
    ----------
    command_config : command configuration object
        Instance of a class that inherits from
        :class:`~gaps.cli.command.AbstractBaseCLICommandConfiguration`
        (i.e. :class:`gaps.cli.command.CLICommandFromClass`,
        :class:`gaps.cli.command.CLICommandFromFunction`, etc.).

    Returns
    -------
    click.command
        A ``click`` command generated from the command configuration.
    """
    doc = command_config.documentation
    name = command_config.name
    params = [
        click.Option(
            param_decls=["--config_file", "-c"],
            required=True,
            type=click.Path(exists=True),
            help=doc.config_help(name),
        )
    ]

    return _WrappedCommand(
        name,
        context_settings=None,
        callback=partial(
            from_config,
            command_config=command_config,
        ),
        params=params,
        help=doc.command_help(name),
        epilog=None,
        short_help=None,
        options_metavar="[OPTIONS]",
        add_help_option=True,
        no_args_is_help=True,
        hidden=False,
        deprecated=False,
    )


def make_cli(commands, info=None):
    """Create a pipeline CLI to split execution across HPC nodes.

    This function generates a CLI for a package based on the input
    command configurations. Each command configuration is based around
    a function that is to be executed on one or more nodes.

    Parameters
    ----------
    commands : list of command configurations
        List of command configs to convert to click commands. See the
        :class:`~gaps.cli.command.CLICommandFromClass` or
        :class:`~gaps.cli.command.CLICommandFromFunction`
        documentation for a description of the input options. Each
        command configuration is converted into a subcommand. Any
        command configuration with ``project_points`` in the
        `split_keys` argument will get a corresponding
        ``collect-{command name}`` command that collects the outputs of
        the spatially-distributed command.
    info : dict, optional
        A dictionary that contains optional info about the calling
        program to include in the CLI. Allowed keys include:

            name : str
                Name of program to display at the top of the CLI help.
                This input is optional, but specifying it yields richer
                documentation for the main CLI command.
            version : str
                Include program version with a ``--version`` CLI option.

        By default, ``None``.

    Returns
    -------
    click.Group
        A group containing the requested subcommands. This can be used
        as the entrypoint to the package CLI.
    """
    info = info or {}
    command_generator = _CLICommandGenerator(commands)
    commands = command_generator.generate()

    options = [
        click.Option(
            param_decls=["-v", "--verbose"],
            is_flag=True,
            help="Flag to turn on debug logging. Default is not verbose.",
        ),
    ]

    prog_name = info.get("name")
    if prog_name:
        main_help = _main_command_help(
            prog_name, command_generator.command_configs
        )
    else:
        main_help = "Command Line Interface"

    main = click.Group(
        help=main_help, params=options, callback=_main_cb, commands=commands
    )
    version = info.get("version")
    if version is not None:
        main = click.version_option(version=version)(main)

    return main


@click.pass_context
def _main_cb(ctx, verbose):
    """Set the obj and verbose settings of the commands"""
    ctx.ensure_object(dict)
    ctx.obj["VERBOSE"] = verbose
    ctx.obj["PIPELINE_STEP"] = ctx.invoked_subcommand
    ctx.max_content_width = 92



"""Legacy reV-like status manager"""

import logging
from abc import abstractmethod

import gaps.batch
import gaps.status
import gaps.pipeline
import gaps.hpc


logger = logging.getLogger(__name__)


def _as_gaps_hardware(hardware):
    """Convert hardware string to gaps-compliant hardware"""
    hardware = hardware.lower()
    if hardware == "pbs":
        hardware = "peregrine"
    return hardware


class PipelineError(Exception):
    """Error for pipeline execution failure"""


class HardwareStatusRetriever(gaps.status.HardwareStatusRetriever):
    """Query hardware for job status"""

    def __init__(self, hardware="slurm", subprocess_manager=None):
        """Initialize `HardwareStatusRetriever`

        Parameters
        ----------
        hardware : str
            Name of hardware that a pipeline is being run on: eagle,
            slurm, local. Defaults to "slurm".
        subprocess_manager : PBS | SLURM | None, optional
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC. By default, `None`.
        """
        super().__init__(subprocess_manager)
        hardware = _as_gaps_hardware(hardware)
        self.hardware = gaps.status._validate_hardware(hardware)  # noqa

    def __getitem__(self, key):
        job_id, __ = key
        if not job_id:
            return None

        if self.subprocess_manager is not None:
            return self.subprocess_manager.check_status_using_job_id(job_id)

        return self.hardware.check_status_using_job_id(job_id)


class Status(gaps.status.Status):
    """Base class for data pipeline health and status information"""

    @classmethod
    def retrieve_job_status(
        cls,
        status_dir,
        module,
        job_name,
        hardware="slurm",
        subprocess_manager=None,
    ):
        """Update and retrieve job status.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        module : str
            Module that the job belongs to.
        job_name : str
            Unique job name identification.
        hardware : str
            Name of hardware that this pipeline is being run on: eagle,
            slurm, local. Defaults to "slurm". This specifies how job
            are queried for status.
        subprocess_manager : None | SLURM
            Optional initialized subprocess manager to use to check job
            statuses. This can be input with cached queue data to avoid
            constantly querying the HPC.

        Returns
        -------
        status : str | None
            Status string or None if job/module not found.
        """
        hardware = _as_gaps_hardware(hardware)
        hsr = HardwareStatusRetriever(hardware, subprocess_manager)
        status = (
            cls(status_dir)._retrieve_job_status(module, job_name, hsr)  # noqa
        )
        if status == gaps.status.StatusOption.NOT_SUBMITTED:
            status = None
        return status

    @classmethod
    def add_job(
        cls, status_dir, module, job_name, replace=False, job_attrs=None
    ):
        """Add a job to status json.

        Parameters
        ----------
        status_dir : str
            Directory containing json status file.
        module : str
            Module that the job belongs to.
        job_name : str
            Unique job name identification.
        replace : bool
            Flag to force replacement of pre-existing job status.
        job_attrs : dict
            Job attributes. Should include 'job_id' if running on HPC.
        """
        cls.mark_job_as_submitted(
            status_dir=status_dir,
            pipeline_step=module,
            job_name=job_name,
            replace=replace,
            job_attrs=job_attrs,
        )

    @staticmethod
    def make_job_file(status_dir, module, job_name, attrs):
        """Make a json file recording the status of a single job.

        Parameters
        ----------
        status_dir : str
            Directory to put json status file.
        module : str
            Module that the job belongs to.
        job_name : str
            Unique job name identification.
        attrs : str
            Dictionary of job attributes that represent the job status
            attributes.
        """
        gaps.status.Status.make_single_job_file(
            status_dir=status_dir,
            pipeline_step=module,
            job_name=job_name,
            attrs=attrs,
        )

    @staticmethod
    def _update_job_status_from_hardware(job_data, hardware_status_retriever):
        """Update job status from HPC hardware if needed"""

        # init defaults in case job/command not in status file yet
        job_status = job_data.get(gaps.status.StatusField.JOB_STATUS, None)
        job_id = job_data.get(gaps.status.StatusField.JOB_ID, None)
        job_hardware = job_data.get(gaps.status.StatusField.HARDWARE, None)

        # get job status from hardware
        current = hardware_status_retriever[job_id, job_hardware]

        # No current status and job was not successful: failed!
        if (
            current is None
            and job_status != gaps.status.StatusOption.SUCCESSFUL
        ):
            job_data[gaps.status.StatusField.JOB_STATUS] = (
                gaps.status.StatusOption.FAILED
            )

        # do not overwrite a successful or failed job status.
        elif (
            current != job_status
            and job_status not in gaps.status.StatusOption.members_as_str()
        ):
            job_data[gaps.status.StatusField.JOB_STATUS] = current


class Pipeline(gaps.pipeline.Pipeline):
    """Legacy reV-like pipeline execution framework"""

    @abstractmethod
    def __init__(self, pipeline, monitor=True, verbose=False):
        """

        When subclassing this object, you MUST set the following
        properties in the __init__ method:

        monitor : bool
            Flag to perform continuous monitoring of the pipeline.
        verbose : bool
            Flag to submit pipeline steps with "-v" flag for debug
            logging.
        _config : object
            reV-style config object. The config object MUST have the
            following attributes:

            name : str
                Job name (typically defaults to the output directory
                name).
            dirout : str
                Output file directory (typically the same directory
                that contains the config file).
            hardware : str
                Name of hardware that the pipeline is being run on
                (typically "eagle").
        _run_list : list
            List of dictionaries, each with a single key-value pair,
            where the key represents the command and the value
            represents the command config filepath to substitute into
            the :attr:`CMD_BASE` string.

        You must also call `self._init_status()` in the initializer.
        If you want logging outputs during the submit step, make sure
        to init the "gaps" logger.

        Parameters
        ----------
        pipeline : str | dict
            Pipeline config file path or dictionary.
        monitor : bool, optional
            Flag to perform continuous monitoring of the pipeline.
            By default, `True`.
        verbose : bool, optional
            Flag to submit pipeline steps with "-v" flag for debug
            logging. By default, `False`.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def CMD_BASE(self):  # noqa: N802
        """str: Formattable string of the base pipeline CLI command"""
        raise NotImplementedError

    @property
    @abstractmethod
    def COMMANDS(self):  # noqa: N802
        """list: List of pipeline command names (as str)"""
        raise NotImplementedError

    @property
    def _out_dir(self):
        """path-like: Output directory"""
        return self._config.dirout  # cspell:disable-line

    @property
    def _name(self):
        """str: Name of the pipeline job (directory of status file)"""
        return self._config.name

    @property
    def _hardware(self):
        """{'local', 'slurm', 'eagle'}: Hardware option"""
        return self._config.hardware

    def _submit(self, step):
        """Submit a step in the pipeline"""

        command, f_config = self._get_command_config(step)
        cmd = self._get_cmd(command, f_config)

        logger.info(
            "Pipeline submitting: %r for job %r", command, self._config.name
        )
        logger.debug("Pipeline submitting subprocess call:\n\t%r", cmd)

        try:
            stderr = gaps.hpc.submit(cmd)[1]
        except OSError:
            logger.exception(
                "Pipeline subprocess submission returned an error"
            )
            raise

        if stderr:
            logger.warning("Subprocess received stderr: \n%s", stderr)

    def _get_command_config(self, step):
        """Get the (command, config) key pair"""
        pipe_step = self._run_list[step]
        return pipe_step.command, pipe_step.config_path

    def _get_cmd(self, command, f_config):
        """Get the python cli call string"""

        if command not in self.COMMANDS:
            msg = (
                f"Could not recognize command {command!r}. "
                f"Available commands are: {self.COMMANDS!r}"
            )
            raise KeyError(msg) from None

        cmd = self.CMD_BASE.format(fp_config=f_config, command=command)
        if self.verbose:
            cmd += " -v"

        return cmd

    @classmethod
    def run(cls, pipeline, monitor=True, verbose=False):
        """Run the reV-style pipeline

        Parameters
        ----------
        pipeline : str | dict
            Pipeline config file path or dictionary.
        monitor : bool, optional
            Flag to perform continuous monitoring of the pipeline.
            By default, `True`.
        verbose : bool, optional
            Flag to submit pipeline steps with "-v" flag for debug
            logging. By default, `False`.
        """

        pipe = cls(pipeline, monitor=monitor, verbose=verbose)
        pipe._main()  # noqa: SLF001


class BatchJob(gaps.batch.BatchJob):
    """Legacy reV-like batch job framework

    To use this class, simply override the following two attributes:

        PIPELINE_CLASS : `Pipeline`
            Pipeline class with at least two class methods: `run` and
            `cancel_all` The `run` method must take `pipeline_config`,
            `monitor`, and `verbose` as arguments.
        PIPELINE_BACKGROUND_METHOD : callable
            Callable to run pipeline in the background with monitoring.
            If you set this as a static method (e.g.
            `PIPELINE_BACKGROUND_METHOD = staticmethod(my_callable)`),
            then this function should take exactly two arguments:
            `pipeline_config` and `verbose`. Otherwise, the function
            should take three arguments, where the first is a reference
            to this batch class, and the last two are the same as above.

    """

    @property
    @abstractmethod
    def PIPELINE_CLASS(self):  # noqa: N802
        """str: Formattable string of the base pipeline CLI command"""
        raise NotImplementedError

    @property
    @abstractmethod
    def PIPELINE_BACKGROUND_METHOD(self):  # noqa: N802
        """str: Formattable string of the base pipeline CLI command"""
        raise NotImplementedError

    def _run_pipelines(self, monitor_background=False, verbose=False):
        """Run the reV pipeline modules for each batch job

        Parameters
        ----------
        monitor_background : bool
            Flag to monitor all batch pipelines continuously in the
            background using the nohup command. Note that the
            stdout/stderr will not be captured, but you can set a
            pipeline "log_file" to capture logs.
        verbose : bool
            Flag to turn on debug logging for the pipelines.
        """

        for sub_directory in self.sub_dirs:
            pipeline_config = sub_directory / self._pipeline_fp.name
            if not pipeline_config.is_file():
                msg = (
                    f"Could not find pipeline config to run: "
                    f"'{pipeline_config.as_posix()}'"
                )
                raise PipelineError(msg)

            pipeline_config = pipeline_config.as_posix()
            if monitor_background:
                self.PIPELINE_BACKGROUND_METHOD(
                    pipeline_config, verbose=verbose
                )
            else:
                self.PIPELINE_CLASS.run(
                    pipeline_config, monitor=False, verbose=verbose
                )

    def _cancel_all(self):
        """Cancel all reV pipeline modules for all batch jobs"""
        for sub_directory in self.sub_dirs:
            pipeline_config = sub_directory / self._pipeline_fp.name
            if pipeline_config.is_file():
                self.PIPELINE_CLASS.cancel_all(pipeline_config.as_posix())

    @classmethod
    def cancel_all(cls, config, verbose=False):  # noqa: ARG003
        """Cancel all reV pipeline modules for all batch jobs

        Parameters
        ----------
        config : str
            File path to batch config json or csv (str).
        verbose : bool
            Flag to turn on debug logging.
        """

        cls(config)._cancel_all()  # noqa: SLF001

    @classmethod
    def delete_all(cls, config, verbose=False):  # noqa: ARG003
        """Delete all reV batch sub job folders based on the job summary
        csv in the batch config directory.

        Parameters
        ----------
        config : str
            File path to batch config json or csv (str).
        verbose : bool
            Flag to turn on debug logging.
        """
        cls(config).delete()

    @classmethod
    def run(
        cls,
        config,
        dry_run=False,
        delete=False,
        monitor_background=False,
        verbose=False,
    ):
        """Run the reV batch job from a config file

        Parameters
        ----------
        config : str
            File path to config json or csv (str).
        dry_run : bool
            Flag to make job directories without running.
        delete : bool
            Flag to delete all batch job sub directories based on the
            job summary csv in the batch config directory.
        monitor_background : bool
            Flag to monitor all batch pipelines continuously
            in the background using the nohup command. Note that the
            stdout/stderr will not be captured, but you can set a
            pipeline "log_file" to capture logs.
        verbose : bool
            Flag to turn on debug logging for the pipelines.
        """

        b = cls(config)
        if delete:
            b.delete()
        else:
            b._make_job_dirs()  # noqa: SLF001
            if not dry_run:
                b._run_pipelines(  # noqa: SLF001
                    monitor_background=monitor_background, verbose=verbose
                )



"""Logging utilities

Some of the code in this module is borrowed from rex
(https://github.com/NREL/rex).
"""

import sys
import logging
from functools import partial
from pathlib import Path

import psutil

import rex
import gaps

FORMAT = "gaps (%(levelname)s) - [%(filename)s:%(lineno)d] : %(message)s"
"""Default format for gaps logging"""


def log_versions():
    """Log package versions"""
    gaps.logger.info("Running with gaps version %s", gaps.__version__)
    rex.utilities.log_versions(gaps.logger)


def log_mem(log_level="DEBUG"):
    """Log the memory usage to the input logger object

    Parameters
    ----------
    logger : logging.Logger
        Logger object to log memory message to.
    log_level : str
        DEBUG or INFO for different log levels for this log message.

    Returns
    -------
    msg : str
        Memory utilization log message string.
    """
    mem = psutil.virtual_memory()
    msg = (
        f"Memory utilization is {mem.used / (1024.0**3):.3f} GB "
        f"out of {mem.total / (1024.0**3):.3f} GB total "
        f"({mem.used / mem.total:.1%} used)"
    )
    gaps.logger.log(logging.getLevelName(log_level), msg)

    return msg


def init_logger(stream=True, level="INFO", file=None, fmt=FORMAT):
    """Initialize and setup logging instance

    Parameters
    ----------
    stream : bool, optional
        Option to add a StreamHandler along with FileHandler.
        By default `True`.
    level : str, optional
        Level of logging to capture. If multiple handlers/log_files are
        requested in a single call of this function, the specified
        logging level will be applied to all requested handlers.
        By default, "INFO".
    file : str | list, optional
        Path to file that should be used for logging. This can also be
        a list of filepaths to use for multiple log files.
        By default, `None`.
    fmt : str, optional
        Format for loggings. By default `FORMAT`.
    """

    file = file or []
    if isinstance(file, str):
        file = [file]

    handlers = [
        make_handler(partial(make_log_file_handler, f), level=level, fmt=fmt)
        for f in file
    ]

    if stream:
        handlers.append(
            make_handler(make_log_stream_handler, level=level, fmt=fmt)
        )

    add_handlers(handlers)


def make_handler(factory, level="INFO", fmt=FORMAT):
    """Make a handler to add to a Logger instance.

    Parameters
    ----------
    factory : callable
        A callable function to create the default handler.
    level : str, optional
        A string representing the logging level for the handler.
        By default "INFO".
    fmt : str, optional
        A string representing the formatting for logging calls.
        By default `FORMAT`.

    Returns
    -------
    handler : `logging.Handler`
        Handler with the specified log level and format.
    """

    handler = factory()

    handler.setLevel(level.upper())

    log_format = logging.Formatter(fmt)
    handler.setFormatter(log_format)

    return handler


def make_log_file_handler(file_path):
    """Make a file handler to add to a Logger instance.

    If the directory structure for `file_path` does not exist, it is
    created before initializing the FileHandler.

    Parameters
    ----------
    file_path : str, optional
        Path to the output log file.

    Returns
    -------
    handler : `logging.FileHandler`
        File handler with `file_path` as the name.
    """

    log_file = Path(file_path).expanduser().resolve()
    log_file.parent.mkdir(parents=True, exist_ok=True)

    handler = logging.FileHandler(log_file, mode="a")
    handler.set_name(log_file.name)

    return handler


def make_log_stream_handler():
    """Make a file handler to add to a Logger instance.

    Returns
    -------
    handler : `logging.StreamHandler`
        Stream handler with name "stream".
    """

    handler = logging.StreamHandler(sys.stdout)
    handler.set_name("stream")

    return handler


def add_handlers(handlers):
    """Add handlers to logger ensuring they do not already exist.

    Parameters
    ----------
    handlers : list
        Handlers to add to logger instance.
    """
    current_handlers = {h.name: h for h in gaps.logger.handlers}
    for handler in handlers:
        name = handler.name
        if name not in current_handlers:
            gaps.logger.addHandler(handler)
            current_handlers.update({name: handler})
        else:
            existing_handler = current_handlers[name]
            if handler.level < existing_handler.level:
                existing_handler.setLevel(handler.level)


def print_logging_info():
    """Print logger names, levels, and handlers"""

    logger_names = [__name__.split(".", maxsplit=1)[0]]
    for name in logger_names:
        print(f"LOGGER: {name!r}")
        logger_to_debug = logging.getLogger(name)
        while logger_to_debug is not None:
            print(
                f"level: {logger_to_debug.level}, "
                f"name: {logger_to_debug.name}, "
                f"handlers: "
            )
            for handler in logger_to_debug.handlers:
                print(f"\t +++ {handler.name} - {handler}")
            logger_to_debug = logger_to_debug.parent


def print_logging_info_all_libraries():
    """Print logger info from all libraries

    Reference
    ---------
    https://stackoverflow.com/questions/3630774/logging-remove-inspect-modify-handlers-configured-by-fileconfig
    """
    loggers = logging.Logger.manager.loggerDict
    for package, logger_ in loggers.items():
        print(f"+ [{package:<40}] {{{__cls_name(logger_)}}} ")
        if isinstance(logger_, logging.PlaceHolder):
            continue
        for handler in gaps.logger.handlers:
            print(f"     +++ {__cls_name(handler)}")


def __cls_name(obj):
    """Format the class name of the input logger/handler object"""
    return str(obj.__class__)[8:-2]



"""CLI documentation utilities"""

from copy import deepcopy
from itertools import chain
from functools import lru_cache
from inspect import signature, isclass

from numpydoc.docscrape import NumpyDocString

from gaps.status import HardwareOption
from gaps.config import config_as_str_for_docstring, ConfigType
from gaps.utilities import _is_sphinx_build


_TAB_LENGTH = 4
DEFAULT_EXEC_VALUES = {
    "option": "local",
    "allocation": "[REQUIRED IF ON HPC]",
    "walltime": "[REQUIRED IF ON HPC]",
    "qos": "normal",
    "memory": None,
    "nodes": 1,
    "queue": None,
    "feature": None,
    "conda_env": None,
    "module": None,
    "sh_script": None,
    "num_test_nodes": None,
}

EXTRA_EXEC_PARAMS = {
    "max_workers": "Maximum number of parallel workers run on each node.",
    "sites_per_worker": "Number of sites to run in series on a worker.",
    "memory_limit_per_worker": "Sets the memory limit *per worker*.",
    "memory_utilization_limit": """Memory utilization limit (fractional).
                Must be a value between 0 and 100. This input sets
                how much data will be stored in-memory at any
                given time before flushing to disk.
    """,
    "timeout": """Number of seconds to wait for parallel run iteration
                to complete before early termination.
    """,
    "pool_size": """Number of futures to submit to a single process pool
                for parallel futures.
    """,
}

CONFIG_TYPES = [
    ConfigType.JSON,
    ConfigType.YAML,
    ConfigType.TOML,
]

MAIN_DOC = """{name} Command Line Interface.

Typically, a good place to start is to set up a {name} job with a pipeline
config that points to several {name} modules that you want to run in serial.

To begin, you can generate some template configuration files using::

    $ {name} template-configs

By default, this generates template JSON configuration files, though you
can request JSON5, YAML, or TOML configuration files instead. You can run
``$ {name} template-configs --help`` on the command line to see all available
options for the ``template-configs`` command. Once the template configuration
files have been generated, you can fill them out by referring to the
module CLI documentation (if available) or the help pages of the module CLIs
for more details on the config options for each CLI command::

    {cli_help_str}

After appropriately filling our the configuration files for each module you
want to run, you can call the {name} pipeline CLI using::

    $ {name} pipeline -c config_pipeline.json

This command will run each pipeline step in sequence.

.. Note:: You will need to re-submit the ``pipeline`` command above after
          each completed pipeline step.

To check the status of the pipeline, you can run::

    $ {name} status

This will print a report to the command line detailing the progress of the
current pipeline. See ``$ {name} status --help`` for all status command
options.

If you need to parameterize the pipeline execution, you can use the ``batch``
command. For details on setting up a batch config file, see the documentation
or run::

    $ {name} batch --help

on the command line. Once you set up a batch config file, you can execute
it using::

    $ {name} batch -c config_batch.json


For more information on getting started, see the
`How to Run a Model Powered by GAPs
<https://nrel.github.io/gaps/misc/examples.users.html>`_ guide.

The general structure of the {name} CLI is given below.
"""

PIPELINE_CONFIG_DOC = """
Path to the ``pipeline`` configuration file. This argument can be
left out, but *one and only one file* with "pipeline" in the
name should exist in the directory and contain the config
information. {sample_config}

Parameters
----------
pipeline : list of dicts
    A list of dictionaries, where each dictionary represents one
    step in the pipeline. Each dictionary should have one of two
    configurations:

        - A single key-value pair, where the key is the name of
          the CLI command to run, and the value is the path to
          a config file containing the configuration for that
          command
        - Exactly two key-value pairs, where one of the keys is
          ``"command"``, with a value that points to the name of
          a command to execute, while the second key is a _unique_
          user-defined name of the pipeline step to execute, with
          a value that points to the path to a config file
          containing the configuration for the command specified
          by the other key. This configuration allows users to
          specify duplicate commands as part of their pipeline
          execution.

logging : dict, optional
    Dictionary containing keyword-argument pairs to pass to
    `init_logger <https://tinyurl.com/47hakp7f/>`_. This
    initializes logging for the submission portion of the
    pipeline. Note, however, that each step (command) will
    **also** record the submission step log output to a
    common "project" log file, so it's only ever necessary
    to use this input if you want a different (lower) level
    of verbosity than the `log_level` specified in the
    config for the step of the pipeline being executed.

"""

BATCH_CONFIG_DOC = """
Path to the ``batch`` configuration file. {sample_config}

Parameters
----------
logging : dict, optional
    Dictionary containing keyword-argument pairs to pass to
    `init_logger <https://tinyurl.com/47hakp7f/>`_. This
    initializes logging for the batch command. Note that
    each pipeline job submitted via batch has it's own
    ``logging`` key that will initialize pipeline step
    logging. Therefore, it's only ever necessary to use
    this input if you want logging information about the
    batching portion of the execution.
pipeline_config : str
    Path to the pipeline configuration defining the commands to
    run for every parametric set.
sets : list of dicts
    A list of dictionaries, where each dictionary defines a
    "set" of parametric runs. Each dictionary should have
    the following keys:

        args : dict
            A dictionary defining the arguments across all input
            configuration files to parameterize. Each argument
            to be parametrized should be a key in this
            dictionary, and the value should be a **list** of the
            parameter values to run for this argument (single-item lists
            are allowed and can be used to vary a parameter value across
            sets).

            {batch_args_dict}

            Remember that the keys in the ``args`` dictionary
            should be part of (at least) one of your other
            configuration files.
        files : list
            A list of paths to the configuration files that
            contain the arguments to be updated for every
            parametric run. Arguments can be spread out over
            multiple files. {batch_files}
        set_tag : str, optional
            Optional string defining a set tag that will prefix
            each job tag for this set. This tag does not need to
            include an underscore, as that is provided during
            concatenation.

"""

_BATCH_ARGS_DICT = """.. tabs::

                .. group-tab:: JSON/JSON5
                    ::

                        {sample_json_args_dict}

                .. group-tab:: YAML
                    ::

                        {sample_yaml_args_dict}

                .. group-tab:: TOML
                    ::

                        {sample_toml_args_dict}


            This example would run a total of six pipelines, one
            with each of the following arg combinations:
            ::

                input_constant_1=18.20, path_to_a_file="/first/path.h5"
                input_constant_1=18.20, path_to_a_file="/second/path.h5"
                input_constant_1=18.20, path_to_a_file="/third/path.h5"
                input_constant_1=19.04, path_to_a_file="/first/path.h5"
                input_constant_1=19.04, path_to_a_file="/second/path.h5"
                input_constant_1=19.04, path_to_a_file="/third/path.h5"


"""
_BATCH_FILES = """For example:

            .. tabs::

                .. group-tab:: JSON/JSON5
                    ::

                        {sample_json_files}

                .. group-tab:: YAML
                    ::

                        {sample_yaml_files}

                .. group-tab:: TOML
                    ::

                        {sample_toml_files}


"""

CONFIG_DOC = """
Path to the ``{name}`` configuration file. {sample_config}

{docstring}

Note that you may remove any keys with a ``null`` value if you do not intend to update them yourself.
"""  # noqa: E501
SAMPLE_CONFIG_DOC = """Below is a sample template config

.. tabs::

    .. group-tab:: JSON/JSON5
        ::

            {template_json_config}

    .. group-tab:: YAML
        ::

            {template_yaml_config}

    .. group-tab:: TOML
        ::

            {template_toml_config}

"""
COMMAND_DOC = """
Execute the ``{name}`` step from a config file.

{desc}

The general structure for calling this CLI command is given below
(add ``--help`` to print help info to the terminal).

"""
EXEC_CONTROL_DOC = """
Parameters
----------
    execution_control : dict
        Dictionary containing execution control arguments. Allowed
        arguments are:

        :option: ({opts})
            Hardware run option. Determines the type of job
            scheduler to use as well as the base AU cost. The
            "slurm" option is a catchall for HPC systems
            that use the SLURM scheduler and **should only be
            used if desired hardware is not listed above**. If
            "local", no other HPC-specific keys in are
            required in `execution_control` (they are ignored
            if provided).
        :allocation: (str)
            HPC project (allocation) handle.
        :walltime: (int)
            Node walltime request in hours.
        :qos: (str, optional)
            Quality-of-service specifier. For Kestrel users:
            This should be one of {{'standby', 'normal',
            'high'}}. Note that 'high' priority doubles the AU
            cost. By default, ``"normal"``.
        :memory: (int, optional)
            Node memory max limit (in GB). By default, ``None``,
            which uses the scheduler's default memory limit.
            For Kestrel users: If you would like to use the
            full node memory, leave this argument unspecified
            (or set to ``None``) if you are running on standard
            nodes. However, if you would like to use the bigmem
            nodes, you must specify the full upper limit of
            memory you would like for your job, otherwise you
            will be limited to the standard node memory size
            (250GB).{n}{eep}
        :queue: (str, optional; PBS ONLY)
            HPC queue to submit job to. Examples include: 'debug',
            'short', 'batch', 'batch-h', 'long', etc.
            By default, ``None``, which uses "test_queue".
        :feature: (str, optional)
            Additional flags for SLURM job (e.g. "-p debug").
            By default, ``None``, which does not specify any
            additional flags.
        :conda_env: (str, optional)
            Name of conda environment to activate. By default,
            ``None``, which does not load any environments.
        :module: (str, optional)
            Module to load. By default, ``None``, which does not
            load any modules.
        :sh_script: (str, optional)
            Extra shell script to run before command call.
            By default, ``None``, which does not run any
            scripts.
        :num_test_nodes: (str, optional)
            Number of nodes to submit before terminating the
            submission process. This can be used to test a
            new submission configuration without submitting
            all nodes (i.e. only running a handful to ensure
            the inputs are specified correctly and the
            outputs look reasonable). By default, ``None``,
            which submits all node jobs.

        Only the `option` key is required for local execution. For
        execution on the HPC, the `allocation` and `walltime` keys are also
        required. All other options are populated with default values,
        as seen above.
    log_directory : str
        Path to directory where logs should be written. Path can be relative
        and does not have to exist on disk (it will be created if missing).
        By default, ``"./logs"``.
    log_level : {{"DEBUG", "INFO", "WARNING", "ERROR"}}
        String representation of desired logger verbosity. Suitable options
        are ``DEBUG`` (most verbose), ``INFO`` (moderately verbose),
        ``WARNING`` (only log warnings and errors), and ``ERROR`` (only log
        errors). By default, ``"INFO"``.

"""
NODES_DOC = (
    "\n        :nodes: (int, optional)"
    "\n            Number of nodes to split the project points across. "
    "\n            Note that the total number of requested nodes for "
    "\n            a job may be larger than this value if the command"
    "\n            splits across other inputs. Default is ``1``."
)
EXTRA_EXEC_PARAM_DOC = "\n        :{name}: ({type})\n            {desc}"


class CommandDocumentation:
    """Generate documentation for a command.

    Commands are typically comprised of one or more functions. This
    definition includes class initializers and object methods.
    Documentation is compiled from all input functions and used to
    generate CLI help docs and template configuration files.
    """

    REQUIRED_TAG = "[REQUIRED]"

    def __init__(self, *functions, skip_params=None, is_split_spatially=False):
        """
        Parameters
        ----------
        *functions : callables
            Functions that comprise a single command for which to
            generate documentation. **IMPORTANT** The extended summary
            will be pulled form the first function only!
        skip_params : set, optional
            Set of parameter names (str) to exclude from documentation.
            Typically this is because the user would not explicitly have
            to specify these. By default, `None`.
        is_split_spatially : bool, optional
            Flag indicating whether or not this function is split
            spatially across nodes. If `True`, a "nodes" option is added
            to the execution control block of the generated
            documentation. By default, `False`.
        """
        self.signatures = [
            signature(func) for func in _as_functions(functions)
        ]
        self.docs = [
            NumpyDocString(func.__doc__ or "")
            for func in _as_functions(functions)
        ]
        self.param_docs = {
            p.name: p for doc in self.docs for p in doc["Parameters"]
        }
        self.skip_params = set() if skip_params is None else set(skip_params)
        self.skip_params |= {"cls", "self"} | set(EXTRA_EXEC_PARAMS)
        self.is_split_spatially = is_split_spatially

    @property
    def default_exec_values(self):
        """dict: Default "execution_control" config"""
        exec_vals = deepcopy(DEFAULT_EXEC_VALUES)
        if not self.is_split_spatially:
            exec_vals.pop("nodes", None)
        for param in EXTRA_EXEC_PARAMS:
            if self._param_in_func_signature(param):
                exec_vals[param] = (
                    self.REQUIRED_TAG
                    if self.param_required(param)
                    else self._param_value(param).default
                )
        return exec_vals

    @property
    def exec_control_doc(self):
        """str: Execution_control documentation"""
        nodes_doc = NODES_DOC if self.is_split_spatially else ""
        hardware_options = str([f"{opt}" for opt in HardwareOption])
        hardware_options = hardware_options.replace("[", "{").replace("]", "}")
        return EXEC_CONTROL_DOC.format(
            opts=hardware_options, n=nodes_doc, eep=self._extra_exec_param_doc
        )

    @property
    def _extra_exec_param_doc(self):
        """str: Docstring formatted with the info from the input func"""
        return "".join(
            [
                self._format_extra_exec_param_doc(param_name)
                for param_name in EXTRA_EXEC_PARAMS
            ]
        )

    def _format_extra_exec_param_doc(self, param_name):
        """Format extra exec control parameters"""
        param = self.param_docs.get(param_name)
        try:
            return EXTRA_EXEC_PARAM_DOC.format(
                name=param_name, type=param.type, desc=" ".join(param.desc)
            )
        except AttributeError:
            pass

        if self._param_in_func_signature(param_name):
            if self.param_required(param_name):
                param_type = "int"
                default_text = ""
            else:
                default_val = self._param_value(param_name).default
                if default_val is None:
                    param_type = "int, optional"
                else:
                    param_type = f"{type(default_val).__name__}, optional"
                default_text = f"By default, ``{default_val}``."
            return EXTRA_EXEC_PARAM_DOC.format(
                name=param_name,
                type=param_type,
                desc=" ".join([EXTRA_EXEC_PARAMS[param_name], default_text]),
            )
        return ""

    @property
    def required_args(self):
        """set: Required parameters of the input function"""
        return {
            name
            for sig in self.signatures
            for name, param in sig.parameters.items()
            if not name.startswith("_")
            and param.default is param.empty
            and name not in self.skip_params
        }

    @property
    def template_config(self):
        """dict: A template configuration file for this function"""
        config = {
            "execution_control": self.default_exec_values,
            "log_directory": "./logs",
            "log_level": "INFO",
        }
        config.update(
            {
                x: self.REQUIRED_TAG if v.default is v.empty else v.default
                for sig in self.signatures
                for x, v in sig.parameters.items()
                if not x.startswith("_") and x not in self.skip_params
            }
        )
        return config

    @property
    def _parameter_npd(self):
        """NumpyDocString: Parameter help `NumpyDocString` instance"""
        param_doc = NumpyDocString("")
        param_doc["Parameters"] = [
            p
            for p in self.param_docs.values()
            if p.name in self.template_config
        ]
        return param_doc

    @property
    def parameter_help(self):
        """str: Parameter help for the func"""
        return str(self._parameter_npd)

    @property
    def hpc_parameter_help(self):
        """str: Function parameter help, including execution control"""
        exec_dict_param = [
            p
            for p in NumpyDocString(self.exec_control_doc)["Parameters"]
            if p.name in {"execution_control", "log_directory", "log_level"}
        ]
        param_doc = deepcopy(self._parameter_npd)
        param_doc["Parameters"] = exec_dict_param + param_doc["Parameters"]
        return "\n".join(_format_lines(str(param_doc).split("\n")))

    @property
    def extended_summary(self):
        """str: Function extended summary, without extra whitespace"""
        return "\n".join(
            _uniform_space_strip(self.docs[0]["Extended Summary"])
        )

    def config_help(self, command_name):
        """Generate a config help string for a command.

        Parameters
        ----------
        command_name : str
            Name of command for which the config help is being
            generated.

        Returns
        -------
        str
            Help string for the config file.
        """
        if _is_sphinx_build():
            sample_config = SAMPLE_CONFIG_DOC.format(
                template_json_config=config_as_str_for_docstring(
                    self.template_config, config_type=ConfigType.JSON
                ),
                template_yaml_config=config_as_str_for_docstring(
                    self.template_config, config_type=ConfigType.YAML
                ),
                template_toml_config=config_as_str_for_docstring(
                    self.template_config, config_type=ConfigType.TOML
                ),
            )
        else:
            sample_config = ""

        doc = CONFIG_DOC.format(
            name=command_name,
            sample_config=sample_config,
            docstring=self.hpc_parameter_help,
        )
        return _cli_formatted(doc)

    def command_help(self, command_name):
        """Generate a help string for a command.

        Parameters
        ----------
        command_name : str
            Name of command for which the help string is being
            generated.

        Returns
        -------
        str
            Help string for the command.
        """
        doc = COMMAND_DOC.format(name=command_name, desc=self.extended_summary)
        return _cli_formatted(doc)

    @lru_cache(maxsize=16)  # noqa: B019
    def _param_value(self, param):
        """Extract parameter if it exists in signature"""
        for sig in self.signatures:
            for name in sig.parameters:
                if name == param:
                    return sig.parameters[param]
        return None

    @lru_cache(maxsize=16)  # noqa: B019
    def _param_in_func_signature(self, param):
        """`True` if `param` is a param of the input function"""
        return self._param_value(param) is not None

    @lru_cache(maxsize=16)  # noqa: B019
    def param_required(self, param):
        """Check whether a parameter is required for the run function

        Parameters
        ----------
        param : str
            Name of parameter to check.

        Returns
        -------
        bool
            ``True`` if `param` is a required parameter of `func`.
        """
        param = self._param_value(param)
        if param is None:
            return False
        return param.default is param.empty


def _main_command_help(prog_name, commands):
    """Generate main command help from commands input"""
    cli_help_str = "\n\n    ".join(
        [f"$ {prog_name} --help"]
        + [f"$ {prog_name} {command.name} --help" for command in commands]
    )
    return MAIN_DOC.format(name=prog_name, cli_help_str=cli_help_str)


def _pipeline_command_help(pipeline_config):  # pragma: no cover
    """Generate pipeline command help from a sample config"""
    if not _is_sphinx_build():
        return _cli_formatted(PIPELINE_CONFIG_DOC.format(sample_config=""))

    template_names = [
        "template_json_config",
        "template_yaml_config",
        "template_toml_config",
    ]
    sample_config = SAMPLE_CONFIG_DOC.format(
        **_format_dict(pipeline_config, template_names)
    )
    doc = PIPELINE_CONFIG_DOC.format(sample_config=sample_config)
    return _cli_formatted(doc)


def _batch_command_help():  # pragma: no cover
    """Generate batch command help from a sample config"""
    if not _is_sphinx_build():
        doc = BATCH_CONFIG_DOC.format(
            sample_config="", batch_args_dict="", batch_files=""
        )
        return _cli_formatted(doc)

    format_inputs = {}
    template_config = {
        "logging": {"log_file": None, "log_level": "INFO"},
        "pipeline_config": CommandDocumentation.REQUIRED_TAG,
        "sets": [
            {
                "args": CommandDocumentation.REQUIRED_TAG,
                "files": CommandDocumentation.REQUIRED_TAG,
                "set_tag": "set1",
            },
            {
                "args": CommandDocumentation.REQUIRED_TAG,
                "files": CommandDocumentation.REQUIRED_TAG,
                "set_tag": "set2",
            },
        ],
    }
    template_names = [
        "template_json_config",
        "template_yaml_config",
        "template_toml_config",
    ]

    format_inputs["sample_config"] = SAMPLE_CONFIG_DOC.format(
        **_format_dict(template_config, template_names)
    )

    sample_args_dict = {
        "args": {
            "input_constant_1": [18.02, 19.04],
            "path_to_a_file": [
                "/first/path.h5",
                "/second/path.h5",
                "/third/path.h5",
            ],
        }
    }
    sample_args_names = [
        "sample_json_args_dict",
        "sample_yaml_args_dict",
        "sample_toml_args_dict",
    ]
    format_inputs["batch_args_dict"] = _BATCH_ARGS_DICT.format(
        **_format_dict(sample_args_dict, sample_args_names, batch_docs=True)
    )

    sample_files = {"files": ["./config_run.yaml", "./config_analyze.json"]}
    sample_files_names = [
        "sample_json_files",
        "sample_yaml_files",
        "sample_toml_files",
    ]
    format_inputs["batch_files"] = _BATCH_FILES.format(
        **_format_dict(sample_files, sample_files_names, batch_docs=True)
    )

    doc = BATCH_CONFIG_DOC.format(**format_inputs)
    return _cli_formatted(doc)


def _format_dict(sample, names, batch_docs=False):  # pragma: no cover
    """Format a sample into a documentation config"""
    configs = {}
    for name, c_type in zip(names, CONFIG_TYPES):
        configs[name] = config_as_str_for_docstring(
            sample,
            config_type=c_type,
            num_spaces=(20 if "json" in name else 24) if batch_docs else 12,
        )
        if batch_docs and "json" in name:
            configs[name] = "\n".join(configs[name].split("\n")[1:-1]).lstrip()
    return configs


def _as_functions(functions):
    """Yield from input, converting all classes to their __init__"""
    for func in functions:
        if isclass(func):
            func = func.__init__  # noqa: PLW2901
        yield func


def _line_needs_newline(line):
    r"""Determine whether a \n should be added to the current line"""
    if any(
        f":{key}:" in line
        for key in chain(DEFAULT_EXEC_VALUES, EXTRA_EXEC_PARAMS)
    ):
        return True
    if not line.startswith("    "):
        return True
    if len(line) <= _TAB_LENGTH:
        return True
    return line[_TAB_LENGTH] == " "


def _format_lines(lines):
    """Format docs into longer lines for easier wrapping in CLI"""
    new_lines = [lines[0]]
    current_line = []
    for line in lines[1:]:
        if _line_needs_newline(line):
            current_line = " ".join(current_line)
            if current_line:
                line = f"{current_line}\n{line}"  # noqa: PLW2901
            new_lines.append(line)
            current_line = []
        else:
            if current_line:
                line = line.lstrip()  # noqa: PLW2901
            current_line.append(line)

    return new_lines


def _cli_formatted(doc):
    """Apply minor formatting changes when displaying to CLI"""
    if not _is_sphinx_build():
        doc = doc.replace("``", "`").replace("{{", "{").replace("}}", "}")
    return doc


def _uniform_space_strip(input_strs):
    """Uniformly strip left-hand whitespace from all strings in list"""
    if not input_strs:
        return input_strs

    input_strs = [x.rstrip() for x in input_strs]
    num_spaces_skip = min(
        len(x) - len(x.lstrip()) if x else float("inf") for x in input_strs
    )
    return [x[num_spaces_skip:] if x else x for x in input_strs]



"""GAPs CLI command for spatially distributed function runs"""

import json
import logging
from copy import deepcopy
from warnings import warn
from pathlib import Path
from itertools import product
from inspect import signature, isclass

import click

from rex.utilities.loggers import init_mult  # cspell:disable-line

from gaps import ProjectPoints
from gaps.status import StatusUpdates, QOSOption, HardwareOption
from gaps.config import load_config
from gaps.log import init_logger
from gaps.cli.execution import kickoff_job
from gaps.cli.documentation import EXTRA_EXEC_PARAMS
from gaps.exceptions import gapsKeyError
from gaps.warn import gapsWarning

logger = logging.getLogger(__name__)

_CMD_LIST = [
    "from gaps.cli.config import run_with_status_updates",
    "from {run_func_module} import {run_func_name}",
    'su_args = "{project_dir}", "{pipeline_step}", "{job_name}"',
    "run_with_status_updates("
    "   {run_func_name}, {node_specific_config}, {logging_options}, su_args, "
    "   {exclude_from_status}"
    ")",
]
TAG = "_j"
MAX_AU_BEFORE_WARNING = {
    "eagle": 10_000,
    "kestrel": 35_000,
}
GAPS_SUPPLIED_ARGS = {
    "tag",
    "command_name",
    "pipeline_step",
    "config_file",
    "project_dir",
    "job_name",
    "out_dir",
    "out_fpath",
    "config",
    "log_directory",
    "verbose",
}


class _FromConfig:
    """Utility class for running a function from a config file"""

    def __init__(self, ctx, config_file, command_config):
        """

        Parameters
        ----------
        ctx : click.Context
            Click context for the invoked command.
        config_file : path-like
            Path to input file containing key-value pairs as input to
            function.
        step : str
            Name of step being run. This will be used to key the status
            dictionary, so it must be unique to the pipeline.
        command_config : `gaps.cli.cli.CLICommandFromFunction`
            A command configuration object containing info such as the
            command name, run function, pre-processing function,
            function documentation, etc.
        """
        self.ctx = ctx
        self.config_file = Path(config_file).expanduser().resolve()
        self.command_config = command_config
        self.config = load_config(config_file)
        self.log_directory = None
        self.verbose = None
        self.exec_kwargs = None
        self.logging_options = None
        self.exclude_from_status = None
        self._include_tag_in_out_fpath = (
            self.command_config.is_split_spatially
            and self.config.get("execution_control", {}).get("nodes", 1) > 1
        )

    @property
    def project_dir(self):
        """`Path`: Path to project directory"""
        return self.config_file.parent

    @property
    def command_name(self):
        """str: Name of command being run"""
        return self.command_config.name

    @property
    def pipeline_step(self):
        """str: Name of pipeline_step being run"""
        return self.ctx.obj.get("PIPELINE_STEP", self.command_name)

    @property
    def job_name(self):
        """str: Name of job being run"""
        return "_".join(
            [self.project_dir.name, self.command_name.replace("-", "_")]
        )

    def enable_logging(self):
        """Enable logging based on config file input"""
        self.log_directory = self.config.pop(
            "log_directory", (self.project_dir / "logs").as_posix()
        )
        self.log_directory = Path(self.log_directory)
        self.verbose = self.config.pop(
            "log_level", "INFO"
        ) == "DEBUG" or self.ctx.obj.get("VERBOSE", False)
        pipe_log_file = self.log_directory / f"{self.job_name}.log"
        init_logger(
            stream=self.ctx.obj.get("LOG_STREAM", True),
            level="DEBUG" if self.verbose else "INFO",
            file=pipe_log_file.as_posix(),
        )
        return self

    def validate_config(self):
        """Validate the user input config file"""
        logger.debug("Validating %r", str(self.config_file))
        _validate_config(self.config, self.command_config.documentation)
        return self

    def preprocess_config(self):
        """Apply preprocessing function to config file"""

        preprocessor_kwargs = {
            "config": self.config,
            "command_name": self.command_name,
            "pipeline_step": self.pipeline_step,
            "config_file": self.config_file,
            "project_dir": self.project_dir,
            "job_name": self.job_name,
            "out_dir": self.project_dir,
            "out_fpath": self.project_dir / self.job_name,
            "log_directory": self.log_directory,
            "verbose": self.verbose,
        }
        extra_preprocessor_kwargs = {
            k: self.config[k]
            for k in self.command_config.preprocessor_args
            if k not in preprocessor_kwargs and k in self.config
        }
        preprocessor_kwargs.update(extra_preprocessor_kwargs)
        preprocessor_defaults = {
            k: v
            for k, v in self.command_config.preprocessor_defaults.items()
            if k not in preprocessor_kwargs
        }
        preprocessor_kwargs.update(preprocessor_defaults)
        preprocessor_kwargs = {
            k: v
            for k, v in preprocessor_kwargs.items()
            if k in self.command_config.preprocessor_args
        }
        self.config = self.command_config.config_preprocessor(
            **preprocessor_kwargs
        )
        return self

    def set_exec_kwargs(self):
        """Extract the execution control dictionary"""
        self.exec_kwargs = {
            "option": "local",
            "sh_script": "",
            "stdout_path": (self.log_directory / "stdout").as_posix(),
        }

        self.exec_kwargs.update(self.config.get("execution_control", {}))
        extra_params = set()
        for extra_exec_param in EXTRA_EXEC_PARAMS:
            if extra_exec_param in self.config:
                extra_params.add(extra_exec_param)
                param = self.config.pop(extra_exec_param)
                self.exec_kwargs[extra_exec_param] = param

        if extra_params:
            msg = (
                f"Found key(s) {extra_params} outside of 'execution_control'. "
                "Moving these keys into 'execution_control' block. To "
                "silence this warning, please specify these keys inside of "
                "the 'execution_control' block."
            )
            warn(msg, gapsWarning)

        return self

    def set_logging_options(self):
        """Assemble the logging options dictionary"""
        self.logging_options = {
            "name": self.job_name,
            "log_directory": self.log_directory.as_posix(),
            "verbose": self.verbose,
            "node": self.exec_kwargs.get("option", "local") != "local",
        }
        return self

    def set_exclude_from_status(self):
        """Assemble the exclusion keyword set"""
        self.exclude_from_status = {"project_points"}
        self.exclude_from_status |= set(
            self.config.pop("exclude_from_status", set())
        )
        self.exclude_from_status = [
            key.lower()
            for key in self.exclude_from_status
            if key.lower() != "tag"
        ]
        return self

    def prepare_context(self):
        """Add required key-val;ue pairs to context object"""
        self.ctx.obj["COMMAND_NAME"] = self.command_name
        self.ctx.obj["PIPELINE_STEP"] = self.pipeline_step
        self.ctx.obj["OUT_DIR"] = self.project_dir
        return self

    def log_job_info(self):
        """Log information about job submission"""
        logger.info(
            "Running %s from config file: %r",
            self.command_name,
            str(self.config_file),
        )
        logger.info("Target output directory: %r", str(self.project_dir))
        logger.info("Target logging directory: %r", str(self.log_directory))
        return self

    def kickoff_jobs(self):
        """Kickoff jobs across nodes based on config and run function"""
        keys_to_run, lists_to_run = self._keys_and_lists_to_run()

        try:
            jobs = sorted(product(*lists_to_run))
        except TypeError:
            jobs = list(product(*lists_to_run))

        self._warn_about_excessive_au_usage(len(jobs))
        extra_exec_args = self._extract_extra_exec_args_for_command()

        for tag, values, exec_kwargs in self._with_tagged_context(jobs):
            node_specific_config = self._compile_node_config(tag)
            node_specific_config.update(extra_exec_args)

            for key, val in zip(keys_to_run, values):
                if isinstance(key, str):
                    node_specific_config.update({key: val})
                else:
                    node_specific_config.update(dict(zip(key, val)))

            cmd = self._compile_run_command(node_specific_config)
            kickoff_job(self.ctx, cmd, exec_kwargs)

        return self

    def _with_tagged_context(self, jobs):
        """Iterate over jobs and populate context with job name"""
        num_jobs_submit = len(jobs)

        exec_kwargs = deepcopy(self.exec_kwargs)
        num_test_nodes = exec_kwargs.pop("num_test_nodes", None)
        if num_test_nodes is None:
            num_test_nodes = float("inf")

        for node_index, values in enumerate(jobs):
            if node_index >= num_test_nodes:
                return

            tag = _tag(node_index, num_jobs_submit)
            self.ctx.obj["NAME"] = f"{self.job_name}{tag}"
            yield tag, values, exec_kwargs

    def _compile_node_config(self, tag):
        """Compile initial node-specific config"""
        job_name = self.ctx.obj["NAME"]
        node_specific_config = deepcopy(self.config)
        node_specific_config.pop("execution_control", None)
        node_specific_config.update(
            {
                "tag": tag,
                "command_name": self.command_name,
                "pipeline_step": self.pipeline_step,
                "config_file": self.config_file.as_posix(),
                "project_dir": self.project_dir.as_posix(),
                "job_name": job_name,
                "out_dir": self.project_dir.as_posix(),
                "out_fpath": self._suggested_stem(job_name).as_posix(),
                "run_method": getattr(self.command_config, "run_method", None),
            }
        )
        return node_specific_config

    def _compile_run_command(self, node_specific_config):
        """Create run command from config and job name"""
        job_name = self.ctx.obj["NAME"]
        cmd = "; ".join(_CMD_LIST).format(
            run_func_module=self.command_config.runner.__module__,
            run_func_name=self.command_config.runner.__name__,
            node_specific_config=as_script_str(node_specific_config),
            project_dir=self.project_dir.as_posix(),
            logging_options=as_script_str(self.logging_options),
            exclude_from_status=as_script_str(self.exclude_from_status),
            pipeline_step=self.pipeline_step,
            job_name=job_name,
        )
        return f"python -c {cmd!r}"

    def _suggested_stem(self, job_name_with_tag):
        """Determine suggested filepath with filename stem"""
        if self._include_tag_in_out_fpath:
            return self.project_dir / job_name_with_tag
        return self.project_dir / self.job_name

    def _extract_extra_exec_args_for_command(self):
        """Dictionary of function args from the exec block"""
        extra_exec_args = {}
        for param in EXTRA_EXEC_PARAMS:
            if param not in self.exec_kwargs:
                continue
            extra_exec_args[param] = self.exec_kwargs.pop(param)
        return extra_exec_args

    def _keys_and_lists_to_run(self):
        """Compile run lists based on `command_config.split_keys`"""
        keys_to_run = []
        lists_to_run = []
        keys = sorted(self.command_config.split_keys, key=_project_points_last)
        for key_group in keys:
            keys_to_run.append(key_group)
            if isinstance(key_group, str):
                lists_to_run.append(self.config.get(key_group) or [None])
            else:
                lists_to_run.append(
                    list(
                        zip(*[self.config.get(k) or [None] for k in key_group])
                    )
                )
        return keys_to_run, lists_to_run

    def _warn_about_excessive_au_usage(self, num_jobs):
        """Warn if max job runtime exceeds AU threshold"""
        try:
            max_walltime_per_job = float(self.exec_kwargs.get("walltime"))
        except (TypeError, ValueError):
            return

        qos = self.exec_kwargs.get("qos") or str(QOSOption.UNSPECIFIED)
        try:
            qos_charge_factor = QOSOption(str(qos)).charge_factor
        except ValueError:
            qos_charge_factor = 1

        hardware = self.exec_kwargs.get("option", "local")
        if hardware.casefold() == HardwareOption.SLURM:
            available_opts = [
                f"{opt}"
                for opt in HardwareOption
                if opt != HardwareOption.SLURM and opt.is_hpc
            ]
            msg = (
                "Detected option='slurm' in execution control. Please do not "
                "use this option unless your HPC is explicitly not supported. "
                f"Available HPC options are: {available_opts}"
            )
            warn(msg, gapsWarning)
            return

        try:
            hardware_charge_factor = HardwareOption(hardware).charge_factor
        except ValueError:
            return

        max_au_usage = int(
            num_jobs
            * float(max_walltime_per_job)
            * qos_charge_factor
            * hardware_charge_factor
        )
        max_au_thresh = MAX_AU_BEFORE_WARNING.get(
            hardware.casefold(), float("inf")
        )
        if max_au_usage > max_au_thresh:
            msg = f"Job may use up to {max_au_usage:,} AUs!"
            warn(msg, gapsWarning)

    def run(self):
        """Run the entire config pipeline"""
        try:
            return (
                self.enable_logging()
                .validate_config()
                .log_job_info()
                .preprocess_config()
                .set_exec_kwargs()
                .set_logging_options()
                .set_exclude_from_status()
                .prepare_context()
                .kickoff_jobs()
            )
        except Exception:
            logger.exception("Encountered error while kicking off jobs")
            raise


@click.pass_context
def from_config(ctx, config_file, command_config, pipeline_step=None):
    """Run command from a config file"""
    if pipeline_step is not None:
        ctx.obj["PIPELINE_STEP"] = pipeline_step
    _FromConfig(ctx, config_file, command_config).run()


def _validate_config(config, documentation):
    """Ensure required keys exist and warn user about extra keys"""
    _ensure_required_args_exist(config, documentation)
    _warn_about_extra_args(config, documentation)


def _ensure_required_args_exist(config, documentation):
    """Make sure that args required for func to run exist in config"""
    missing = {
        name for name in documentation.required_args if name not in config
    }

    missing = _mark_extra_exec_params_missing_if_needed(
        documentation, config, missing
    )

    if any(missing):
        msg = (
            f"The following required keys are missing from the configuration "
            f"file: {missing}"
        )
        raise gapsKeyError(msg)


def _mark_extra_exec_params_missing_if_needed(documentation, config, missing):
    """Add extra exec params as missing if not in `execution_control"""
    exec_control = config.get("execution_control", {})
    for param in EXTRA_EXEC_PARAMS:
        param_required = documentation.param_required(param)
        param_missing = param not in config and param not in exec_control
        if param_required and param_missing:
            missing |= {param}
    return missing


def _warn_about_extra_args(config, documentation):
    """Warn user about extra unused keys in the config file"""
    extra = {name for name in config if not _param_in_sig(name, documentation)}
    extra -= {"execution_control", "project_points_split_range"}
    if any(extra):
        msg = (
            "Found unused keys in the configuration file: %s. To silence "
            "this warning, please remove these keys from the input "
            "configuration file."
        )
        warn(msg % extra, gapsWarning)


def _param_in_sig(param, documentation):
    """Determine if ``name`` is an argument in any func signatures"""
    return any(
        param in _public_args(signature)
        for signature in documentation.signatures
    )


def _public_args(func_signature):
    """Gather set of all "public" function args"""
    return {
        param
        for param in func_signature.parameters
        if not param.startswith("_")
    }


def _project_points_last(key):
    """Sorting that always puts "project_points_split_range" last"""
    if isinstance(key, str):
        if key.casefold() == "project_points_split_range":
            return (chr(0x10FFFF),)  # PEP 393
        return (key,)
    return key


def _tag(node_index, num_jobs):
    """Determine node tag based on total number of jobs"""
    n_zfill = len(str(max(0, num_jobs - 1)))
    if num_jobs > 1:
        return f"{TAG}{str(node_index).zfill(n_zfill)}"
    return ""


def as_script_str(input_):
    """Convert input to how it would appear in a python script.

    Essentially this means the input is dumped to json format with some
    minor replacements (e.g. null -> None, etc.). Importantly, all
    string inputs are wrapped in double quotes.

    Parameters
    ----------
    input_ : obj
        Any object that can be serialized with :func:`json.dumps`.

    Returns
    -------
    str
        Input object as it would appear in a python script.

    Examples
    --------
    >>> as_script_str("a")
    "a"

    >>> as_script_str({"a": None, "b": True, "c": False, "d": 3,
    ...                "e": [{"t": "hi"}]})
    {"a": None, "b": True, "c": False, "d": 3, "e": [{"t": "hi"}]}
    """
    return (
        json.dumps(input_)
        .replace("null", "None")
        .replace("false", "False")
        .replace("true", "True")
    )


def run_with_status_updates(
    run_func, config, logging_options, status_update_args, exclude
):
    """Run a function and write status updated before/after execution.

    Parameters
    ----------
    run_func : callable
        A function to run.
    config : dict
        Dictionary of node-specific inputs to `run_func`.
    logging_options : dict
        Dictionary of logging options containing at least the following
        key-value pairs:

            name : str
                Job name; name of log file.
            log_directory : path-like
                Path to log file directory.
            verbose : bool
                Option to turn on debug logging.
            node : bool
                Flag for whether this is a node-level logger. If this is
                a node logger, and the log level is info, the log_file
                will be `None` (sent to stdout).

    status_update_args : iterable
        An iterable containing the first three initializer arguments for
        :class:`StatusUpdates`.
    exclude : collection | None
        A collection (list, set, dict, etc.) of keys that should be
        excluded from the job status file that is written before/after
        the function runs.
    """

    # initialize loggers for multiple modules
    init_mult(  # cspell:disable-line
        logging_options["name"],
        logging_options["log_directory"],
        modules=[run_func.__module__.split(".")[0], "gaps", "rex"],
        verbose=logging_options["verbose"],
        node=logging_options["node"],
    )

    run_kwargs = node_kwargs(run_func, config)
    exclude = exclude or set()
    job_attrs = {
        key: value for key, value in run_kwargs.items() if key not in exclude
    }
    status_update_args = *status_update_args, job_attrs
    with StatusUpdates(*status_update_args) as status:
        out = run_func(**run_kwargs)
        if method := config.get("run_method"):
            func = getattr(out, method)
            run_kwargs = node_kwargs(func, config)
            status.job_attrs.update(
                {
                    key: value
                    for key, value in run_kwargs.items()
                    if key not in exclude
                }
            )
            out = func(**run_kwargs)
        status.out_file = out


def node_kwargs(run_func, config):
    """Compile the function inputs arguments for a particular node.

    Parameters
    ----------
    run_func : callable
        A function to run.
    config : dict
        Dictionary of node-specific inputs to `run_func`.
    logging_options : dict
        Dictionary of logging options containing at least the following
        key-value pairs:

            name : str
                Job name; name of log file.
            log_directory : path-like
                Path to log file directory.
            verbose : bool
                Option to turn on debug logging.
            node : bool
                Flag for whether this is a node-level logger. If this is
                a node logger, and the log level is info, the log_file
                will be `None` (sent to stdout).


    Returns
    -------
    dict
        Function run kwargs to be used on this node.
    """

    split_range = config.pop("project_points_split_range", None)
    if split_range is not None:
        config["project_points"] = ProjectPoints.from_range(
            split_range, config["project_points"]
        )

    sig = signature(run_func)
    run_kwargs = {k: v for k, v in config.items() if k in sig.parameters}
    verb = "Initializing" if isclass(run_func) else "Running"
    logger.debug("%s %r with kwargs: %s", verb, run_func.__name__, run_kwargs)
    return run_kwargs



"""GAPs script CLI function"""

import os
import logging

from gaps.hpc import submit


logger = logging.getLogger(__name__)


def script(_cmd, project_dir):
    """Run a command or script as part of a pipeline step

    This command runs one or more terminal commands/scripts as part of a
    pipeline step.

    Parameters
    ----------
    _cmd : str
        String representation of a command to execute on a node.

    Returns
    -------
    str
        Path to HDF5 file with the collected outputs.
    """
    original_directory = os.getcwd()  # noqa: PTH109
    try:
        os.chdir(project_dir)
        stdout, stderr = submit(_cmd)
        if stdout:
            logger.info("Subprocess received stdout: \n%s", stdout)
        if stderr:
            logger.warning("Subprocess received stderr: \n%s", stderr)
    finally:
        os.chdir(original_directory)



"""GAPs CLI entry points"""

import click

from gaps.cli.status import status_command
from gaps._version import __version__


@click.group()
@click.version_option(version=__version__)
@click.pass_context
def main(ctx):
    """GAPs command line interface"""
    ctx.ensure_object(dict)


main.add_command(status_command(), name="status")


if __name__ == "__main__":
    main(obj={})



"""GAPs Documentation"""

import logging
from pathlib import Path

from .collection import Collector
from .pipeline import Pipeline
from .project_points import ProjectPoints
from .status import Status
from ._version import __version__


GAPS_DIR = Path(__file__).parent
REPO_NAME = __name__
TEST_DATA_DIR = GAPS_DIR.parent / "tests" / "data"

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())
logger.setLevel("DEBUG")
logger.propagate = False



"""GAPs config functions and classes"""

import logging
import collections
from pathlib import Path
from abc import ABC, abstractmethod

import json
import yaml
import toml
import pyjson5

from gaps.log import init_logger
from gaps.utilities import CaseInsensitiveEnum, resolve_path
from gaps.exceptions import gapsValueError

logger = logging.getLogger(__name__)
_CONFIG_HANDLER_REGISTRY = {}


class _JSON5Formatter:
    """Format input JSON5 data with indentation"""

    def __init__(self, data):
        self.data = data

    def _format_as_json(self):
        """Format the data input with as string with indentation"""
        return json.dumps(self.data, indent=4)


class Handler(ABC):
    """ABC for configuration file handler"""

    def __init_subclass__(cls):
        super().__init_subclass__()
        if isinstance(cls.FILE_EXTENSION, str):
            _CONFIG_HANDLER_REGISTRY[cls.FILE_EXTENSION] = cls
        else:
            for file_extension in cls.FILE_EXTENSION:
                _CONFIG_HANDLER_REGISTRY[file_extension] = cls

    @classmethod
    def load(cls, file_name):
        """Load the file contents"""
        config_str = Path(file_name).read_text(encoding="utf-8")
        return cls.loads(config_str)

    @classmethod
    def write(cls, file_name, data):
        """Write the data to a file"""
        with Path(file_name).open("w", encoding="utf-8") as config_file:
            cls.dump(data, config_file)

    @classmethod
    @abstractmethod
    def dump(cls, config, stream):
        """Write the config to a stream (file)"""

    @classmethod
    @abstractmethod
    def dumps(cls, config):
        """Convert the config to a string"""

    @classmethod
    @abstractmethod
    def loads(cls, config_str):
        """Parse the string into a config dictionary"""

    @property
    @abstractmethod
    def FILE_EXTENSION(self):  # noqa: N802
        """str: Enum name to use"""


class JSONHandler(Handler):
    """JSON config file handler"""

    FILE_EXTENSION = "json"

    @classmethod
    def dump(cls, config, stream):
        """Write the config to a stream (JSON file)"""
        return json.dump(config, stream, indent=4)

    @classmethod
    def dumps(cls, config):
        """Convert the config to a JSON string"""
        return json.dumps(config, indent=4)

    @classmethod
    def loads(cls, config_str):
        """Parse the JSON string into a config dictionary"""
        return json.loads(config_str)


class JSON5Handler(Handler):
    """JSON5 config file handler"""

    FILE_EXTENSION = "json5"

    @classmethod
    def dump(cls, config, stream):
        """Write the config to a stream (JSON5 file)"""
        return pyjson5.encode_io(
            _JSON5Formatter(config),
            stream,
            supply_bytes=False,
            tojson="_format_as_json",
        )

    @classmethod
    def dumps(cls, config):
        """Convert the config to a JSON5 string"""
        return pyjson5.encode(
            _JSON5Formatter(config),
            tojson="_format_as_json",
        )

    @classmethod
    def loads(cls, config_str):
        """Parse the JSON5 string into a config dictionary"""
        return pyjson5.decode(config_str, maxdepth=-1)


class YAMLHandler(Handler):
    """YAML config file handler"""

    FILE_EXTENSION = "yaml", "yml"

    @classmethod
    def dump(cls, config, stream):
        """Write the config to a stream (YAML file)"""
        return yaml.safe_dump(config, stream, indent=2, sort_keys=False)

    @classmethod
    def dumps(cls, config):
        """Convert the config to a YAML string"""
        return yaml.safe_dump(config, indent=2, sort_keys=False)

    @classmethod
    def loads(cls, config_str):
        """Parse the YAML string into a config dictionary"""
        return yaml.safe_load(config_str)


class TOMLHandler(Handler):
    """TOML config file handler"""

    FILE_EXTENSION = "toml"

    @classmethod
    def dump(cls, config, stream):
        """Write the config to a stream (TOML file)"""
        return toml.dump(config, stream)

    @classmethod
    def dumps(cls, config):
        """Convert the config to a TOML string"""
        return toml.dumps(config)

    @classmethod
    def loads(cls, config_str):
        """Parse the TOML string into a config dictionary"""
        return toml.loads(config_str)


class _ConfigType(CaseInsensitiveEnum):
    """Base config type enum class only meant to be initialized once"""

    @classmethod
    def _new_post_hook(cls, obj, value):
        """Hook for post-processing after __new__; adds methods"""
        obj.dump = _CONFIG_HANDLER_REGISTRY[value].dump
        obj.dumps = _CONFIG_HANDLER_REGISTRY[value].dumps
        obj.load = _CONFIG_HANDLER_REGISTRY[value].load
        obj.loads = _CONFIG_HANDLER_REGISTRY[value].loads
        obj.write = _CONFIG_HANDLER_REGISTRY[value].write
        return obj


ConfigType = _ConfigType(
    "ConfigType",
    {
        config_type.upper(): config_type
        for config_type in _CONFIG_HANDLER_REGISTRY
    },
)
"""An enumeration of the available gaps config types. """


def config_as_str_for_docstring(
    config, config_type=ConfigType.JSON, num_spaces=12
):
    """Convert a config into a string to be used within a docstring.

    In particular, the config is serialized and extra whitespace is
    added after each newline.

    Parameters
    ----------
    config : dict
        Dictionary containing the configuration to be converted into
        docstring format.
    config_type : :class:`ConfigType`, optional
        A :class:`ConfigType` enumeration value specifying what type
        of config file to generate. By default, :attr:`ConfigType.JSON`.
    num_spaces : int, optional
        Number of spaces to add after a newline. By default, `12`.

    Returns
    -------
    str
        A string version of the input config, conforming to the
        specified config type.
    """
    newline_str = "".join(["\n"] + [" "] * num_spaces)
    return config_type.dumps(config).replace("\n", newline_str)


def load_config(config_filepath, resolve_paths=True):
    """Load a config file

    Parameters
    ----------
    config_filepath : path-like
        Path to config file.
    resolve_paths : bool, optional
        Option to (recursively) resolve file-paths in the dictionary
        w.r.t the config file directory.
        By default, ``True``.

    Returns
    -------
    dict
        Dictionary containing configuration parameters.

    Raises
    ------
    gapsValueError
        If input `config_filepath` has no file ending.
    """
    # TODO: maybe also have a "required keys" argument
    config_filepath = Path(config_filepath).expanduser().resolve()
    if "." not in config_filepath.name:
        msg = (
            f"Configuration file must have a file-ending. Got: "
            f"{config_filepath.name}"
        )
        raise gapsValueError(msg)
    config_type = ConfigType(config_filepath.name.split(".")[-1])
    config = config_type.load(config_filepath)
    if resolve_paths:
        return resolve_all_paths(config, config_filepath.parent)

    return config


def resolve_all_paths(container, base_dir):
    """Perform a deep string replacement and path resolve in `container`

    Parameters
    ----------
    container : dict | list
        Container like a dictionary or list that may (or may not)
        contain relative paths to resolve.
    base_dir : path-like
        Base path to directory from which to resolve path string
        (typically current directory)

    Returns
    -------
    container
        Input container with updated strings.
    """

    if isinstance(container, str):
        # `resolve_path` is safe to call on any string,
        # even if it is not a path
        container = resolve_path(container, Path(base_dir))

    elif isinstance(container, collections.abc.Mapping):
        container = {
            key: resolve_all_paths(val, Path(base_dir))
            for key, val in container.items()
        }

    elif isinstance(container, collections.abc.Sequence):
        container = [
            resolve_all_paths(item, Path(base_dir)) for item in container
        ]

    return container


def init_logging_from_config_file(config_file, background=False):
    """Init logging, taking care of legacy rex-style kwargs.

    Parameters
    ----------
    config_file : path-like
        Path to a config file parsable as a dictionary which may or may
        not contain a "logging" key. If key not found, no further action
        is taken. If key is found, the value is expected to be a
        dictionary of keyword-argument pairs
        to :func:`gaps.log.init_logger`. rex-style keys ("log_level",
        "log_file", "log_format") are allowed. The values of these
        inputs are used to initialize a gaps logger.
    background : bool, optional
        Optional flag to specify background job initialization. If
        ``True``, then stream output is disabled. By default, ``False``.
    """
    config = load_config(config_file)
    if "logging" not in config:
        return

    kwargs = config["logging"]
    kwarg_map = {"log_level": "level", "log_file": "file", "log_format": "fmt"}
    for legacy_kwarg, new_kwarg in kwarg_map.items():
        if legacy_kwarg in kwargs:
            kwargs[new_kwarg] = kwargs.pop(legacy_kwarg)

    if background:
        kwargs["stream"] = False

    init_logger(**kwargs)



"""GAPs collection CLI entry points"""

import glob
import logging
from warnings import warn

from rex import Resource
from gaps import Collector
from gaps.warn import gapsWarning

logger = logging.getLogger(__name__)


def collect(
    _out_path,
    _pattern,
    project_points=None,
    datasets=None,
    purge_chunks=False,
    clobber=True,
):
    """Run collection on local worker.

    Collect data generated across multiple nodes into a single HDF5
    file.

    Parameters
    ----------
    project_points : str | list, optional
        This input should represent the project points that correspond
        to the *full collection* of points contained in the input HDF5
        files to be collected. You may simply point to a ProjectPoints
        csv file that contains the GID's that should be collected. You
        may also input the GID's as a list, though this may not be
        suitable for collections with a large number of points. You may
        also set this to input to ``None`` to generate a list of GID's
        automatically from the input files. By default, `None`.
    datasets : list of str, optional
        List of dataset names to collect into the output file. If
        collection is performed into multiple files (i.e. multiple input
        patterns), this list can contain all relevant datasets across
        all files (a warning wil be thrown, but it is safe to ignore
        it). If ``None``, all datasets from the input files are
        collected. By default, ``None``.
    purge_chunks : bool, optional
        Option to delete single-node input HDF5 files. Note that the
        input files will **not** be removed if any of the datasets they
        contain have not been collected, regardless of the value of this
        input. By default, ``False``.
    clobber : bool, optional
        Flag to purge all collection output HDF5 files prior to running
        the collection step if they exist on disk. This helps avoid any
        surprising data byproducts when re-running the collection step
        in a project directory. By default, ``True``.

    Returns
    -------
    str
        Path to HDF5 file with the collected outputs.
    """
    if "*" not in _pattern:
        logger.info("Collect pattern has no wildcard! No collection performed")
        return str(_out_path)

    logger.info(
        "Collection is being run with collection pattern: %s. Target output "
        "path is: %s",
        _pattern,
        _out_path,
    )

    datasets = _find_datasets(datasets, _pattern)
    collector = Collector(_out_path, _pattern, project_points, clobber=clobber)
    for dataset_name in datasets:
        logger.debug("Collecting %r...", dataset_name)
        collector.collect(dataset_name)

    if purge_chunks:
        collector.purge_chunks()
    else:
        collector.move_chunks()

    return str(_out_path)


def _find_datasets(datasets, pattern):
    """Find datasets from a sample file"""

    with Resource(glob.glob(pattern)[0]) as res:  # noqa: PTH207
        if datasets is None:
            return [
                d
                for d in res
                if not d.startswith("time_index") and d != "meta"
            ]
        missing = {d for d in datasets if d not in res}

    if any(missing):
        msg = (
            f"Could not find the following datasets in the output files: "
            f"{missing}. Skipping..."
        )
        warn(msg, gapsWarning)

    return list(set(datasets) - missing)



"""GAPs Pipeline architecture"""

import time
import logging
from pathlib import Path
from warnings import warn

from gaps.status import Status, StatusOption, StatusField, HardwareOption
from gaps.utilities import recursively_update_dict
from gaps.config import load_config
from gaps.exceptions import (
    gapsConfigError,
    gapsExecutionError,
    gapsKeyError,
    gapsValueError,
)
from gaps.warn import gapsWarning


logger = logging.getLogger(__name__)


class PipelineStep:
    """A Pipeline Config step"""

    COMMAND_KEY = "command"
    _KEYS_PER_STEP = 2

    def __init__(self, step_dict):
        self.name = self.config_path = self._command = None
        self._parse_step_dict(step_dict.copy())

    def _parse_step_dict(self, step_dict):
        """Parse the input step dictionary into name, command, and fp"""
        if len(step_dict) > self._KEYS_PER_STEP:
            msg = (
                f"Pipeline step dictionary can have at most two keys. Got: "
                f"{step_dict}"
            )
            raise gapsConfigError(msg)

        if len(step_dict) > 1 and self.COMMAND_KEY not in step_dict:
            msg = (
                f"The only extra key allowed in pipeline step dictionary "
                f"is {self.COMMAND_KEY!r}. Got dictionary: {step_dict}"
            )
            raise gapsConfigError(msg)

        self._command = step_dict.pop(self.COMMAND_KEY, None)
        self.name, self.config_path = next(iter(step_dict.items()))

    @property
    def command(self):
        """Pipeline command to call"""
        if self._command is None:
            self._command = self.name
        return self._command


class Pipeline:
    """gaps pipeline execution framework"""

    COMMANDS = {}

    def __init__(self, pipeline, monitor=True):
        """

        Parameters
        ----------
        pipeline : path-like
            Pipeline config file path.
        monitor : bool, optional
            Flag to perform continuous monitoring of the pipeline.
            By default, ``True``.

        Raises
        ------
        gapsConfigError
            If "pipeline" key not in config file.
        """
        self.monitor = monitor
        self._out_dir = Path(pipeline).expanduser().parent.resolve()
        self._out_dir = self._out_dir.as_posix()

        config = load_config(pipeline)
        self._run_list = _check_pipeline(config)
        self._init_status()

    def _init_status(self):
        """Initialize the status json in the output directory"""
        status = self.status
        self._run_list = [
            PipelineStep(step_dict) for step_dict in self._run_list
        ]
        for pipe_index, pipe_step in enumerate(self._run_list):
            step_dict = {
                pipe_step.name: {StatusField.PIPELINE_INDEX: pipe_index}
            }
            status.data = recursively_update_dict(status.data, step_dict)

        _dump_sorted(status)

    @property
    def status(self):
        """:class:`~gaps.status.Status`: A pipeline status object"""
        return Status(self._out_dir).update_from_all_job_files(purge=False)

    @property
    def name(self):
        """str: Name of the pipeline job (directory of status file)"""
        return self.status.name

    def _cancel_all_jobs(self):
        """Cancel all jobs in this pipeline"""
        status = self.status
        for job_id, hardware in zip(status.job_ids, status.job_hardware):
            if job_id is None:
                continue

            manager = HardwareOption(hardware).manager
            if manager is None:
                continue

            manager.cancel(job_id)

        logger.info("Pipeline job %r cancelled.", self.name)

    def _main(self):
        """Submitting run steps while monitoring status"""

        for step, pipe_step in enumerate(self._run_list):
            step_status = self._status(step)

            if step_status == StatusOption.SUCCESSFUL:
                logger.debug("Successful: %r.", pipe_step.name)
                continue

            # the submit function handles individual job success/failure
            self._submit(step)

            # do not enter while loop for continuous monitoring
            if not self.monitor:
                break

            time.sleep(1)
            try:
                self._monitor(step)
            except Exception:
                self._cancel_all_jobs()
                raise

        else:
            logger.info("Pipeline job %r is complete.", self.name)
            logger.debug("Output directory is: %r", self._out_dir)

    def _monitor(self, step, seconds=5, step_status=StatusOption.RUNNING):
        """Continuously monitor job until success or failure"""

        while step_status.is_processing:
            time.sleep(seconds)
            HardwareOption.reset_all_cached_queries()
            step_status = self._status(step)

            if step_status == StatusOption.FAILED:
                pipe_step = self._run_list[step]
                msg = (
                    f"Pipeline failed at step {step}: {pipe_step.name!r} "
                    f"for {pipe_step.config_path!r}"
                )
                raise gapsExecutionError(msg)

    def _submit(self, step):
        """Submit a step in the pipeline"""

        pipe_step = self._run_list[step]
        if pipe_step.command not in self.COMMANDS:
            msg = (
                f"Could not recognize command {pipe_step.command!r}. "
                f"Available commands are: {set(self.COMMANDS)!r}"
            )
            raise gapsKeyError(msg) from None

        self.COMMANDS[pipe_step.command].callback(
            pipe_step.config_path, pipeline_step=pipe_step.name
        )

    def _status(self, step):
        """Get a pipeline step status"""

        pipe_step = self._run_list[step]
        status = self.status
        submitted = _check_jobs_submitted(status, pipe_step.name)
        if not submitted:
            return StatusOption.NOT_SUBMITTED

        return self._get_step_return_code(status, pipe_step.name)

    def _get_step_return_code(self, status, step_name):  # noqa: C901
        """Get a return code for a pipeline step based on status object

        Note that it is assumed a job has been submitted before this
        function is called, otherwise the return values make no sense!
        """

        # initialize return code array
        arr = []
        check_failed = False
        status.update_from_all_job_files(check_hardware=True)

        if step_name not in status.data:
            # assume running
            arr = [StatusOption.RUNNING]
        else:
            for job_name, job_info in status.data[step_name].items():
                if job_name == StatusField.PIPELINE_INDEX:
                    continue

                job_status = job_info.get(StatusField.JOB_STATUS)

                if job_status == "successful":
                    arr.append(StatusOption.SUCCESSFUL)
                elif job_status == "failed":
                    arr.append(StatusOption.FAILED)
                    check_failed = True
                elif job_status == "submitted":
                    arr.append(StatusOption.SUBMITTED)
                elif job_status == "not submitted":
                    arr.append(StatusOption.NOT_SUBMITTED)
                elif job_status == "running":
                    arr.append(StatusOption.RUNNING)
                elif job_status is None:
                    arr.append(StatusOption.COMPLETE)
                else:
                    msg = f"Job status code {job_status!r} not understood!"
                    raise gapsValueError(msg)

            _dump_sorted(status)

        return_code = _parse_code_array(arr)

        fail_str = ""
        if return_code != StatusOption.FAILED and check_failed:
            fail_str = ", but some jobs have failed"
        logger.info(
            "Pipeline step %r for job %r %s%s. (%s)",
            step_name,
            self.name,
            return_code.with_verb,
            fail_str,
            time.ctime(),
        )

        return return_code

    @classmethod
    def cancel_all(cls, pipeline):
        """Cancel all jobs corresponding to pipeline

        Parameters
        ----------
        pipeline : path-like
            Pipeline config file path.
        """
        cls(pipeline)._cancel_all_jobs()  # noqa: SLF001

    @classmethod
    def run(cls, pipeline, monitor=True):
        """Run the pipeline

        Parameters
        ----------
        pipeline : path-like
            Pipeline config file path.
        monitor : bool
            Flag to perform continuous monitoring of the pipeline.
        """
        cls(pipeline, monitor=monitor)._main()  # noqa: SLF001


def _check_pipeline(config):
    """Check pipeline steps input"""

    if "pipeline" not in config:
        msg = 'Could not find required key "pipeline" in the pipeline config.'
        raise gapsConfigError(msg)

    pipeline = config["pipeline"]

    if not isinstance(pipeline, list):
        msg = (
            'Config arg "pipeline" must be a list of '
            f"{{command: f_config}} pairs, but received {type(pipeline)}."
        )
        raise gapsConfigError(msg)

    step_names = set()
    duplicate_names = []
    for pipe_step in pipeline:
        pipe_step = PipelineStep(pipe_step)  # noqa: PLW2901

        if pipe_step.name in step_names:
            duplicate_names.append(pipe_step.name)
        step_names.add(pipe_step.name)

        if not Path(pipe_step.config_path).expanduser().resolve().exists():
            msg = (
                "Pipeline step depends on non-existent "
                f"file: {pipe_step.config_path}"
            )
            raise gapsConfigError(msg)

    if duplicate_names:
        msg = (
            f"Pipeline contains duplicate step names: {duplicate_names}. "
            "Please specify unique step names for all steps (use the "
            "'command' key to specify duplicate commands to execute)"
        )
        raise gapsConfigError(msg)

    return pipeline


def _parse_code_array(arr):
    """Parse array of return codes to get single return code for step"""

    # check to see if all have completed, or any have failed
    all_successful = all(status == StatusOption.SUCCESSFUL for status in arr)
    all_completed = not any(status.is_processing for status in arr)
    any_failed = any(status == StatusOption.FAILED for status in arr)
    any_submitted = any(status == StatusOption.SUBMITTED for status in arr)

    # only return success if all have succeeded.
    if all_successful:
        return StatusOption.SUCCESSFUL
    # Only return failed when all have finished.
    if all_completed and any_failed:
        return StatusOption.FAILED
    # only return complete when all have completed
    # (but some should have succeeded or failed)
    if all_completed:
        return StatusOption.COMPLETE
    # only return "running" if all jobs running, else return "submitted"
    if any_submitted:
        return StatusOption.SUBMITTED
    # otherwise, all jobs are still running
    return StatusOption.RUNNING


def _check_jobs_submitted(status, step_name):
    """Check whether jobs have been submitted for a pipeline step"""
    return any(
        job != StatusField.PIPELINE_INDEX for job in status.data.get(step_name)
    )


def _dump_sorted(status):
    """Dump status dict after sorting on PIPELINE_INDEX"""
    pi_key = StatusField.PIPELINE_INDEX

    def _sort_key(status_entry):
        """Sort on pipeline index and key name; non-pipeline at top"""
        try:
            pipe_index = status[status_entry].get(pi_key, -1)
        except AttributeError:
            pipe_index = -1
        return pipe_index, status_entry

    sorted_keys = sorted(status, key=_sort_key)
    status.data = {k: status[k] for k in sorted_keys}
    status.dump()


def parse_previous_status(status_dir, command, key=StatusField.OUT_FILE):
    """Parse key from job status(es) from the previous pipeline step

    Parameters
    ----------
    status_dir : path-like
        Directory containing the status file to parse.
    command : str
        Current CLI command (i.e. current pipeline step).
    key : `StatusField` | str, optional
        Parsing target of previous command. By default,
        `StatusField.OUT_FILE`.

    Returns
    -------
    out : list
        Arguments parsed from the status file in status_dir from
        the command preceding the input command arg. This list is
        empty if the `key` is not found in the job status, or if
        the previous step index does not exist in status.

    Raises
    ------
    gapsKeyError
        If ``command`` not in status.

    Warnings
    --------
    This command **DOES NOT** check the HPC queue for jobs and therefore
    **DOES NOT** update the status of previously running jobs that have
    errored out of the HPC queue. For best results, ensure that all
    previous steps of a pipeline have finished processing before calling
    this function.

    This command will not function properly for pipelines with duplicate
    command calls (i.e. multiple collect calls under different names,
    etc.).
    """

    status = Status(status_dir).update_from_all_job_files(purge=False)

    if (
        command not in status
        or StatusField.PIPELINE_INDEX not in status[command]
    ):
        msg = (
            f"Could not parse data for command {command!r} from status "
            f"file in {status_dir!r}"
        )
        raise gapsKeyError(msg)

    index = int(status.data[command][StatusField.PIPELINE_INDEX]) - 1

    if index < 0:
        index = 0
        msg = (
            f"CLI command {command!r} is attempting to parse a previous "
            f"pipeline step, but it appears to be the first step. "
            f"Attempting to parse data from {command!r}."
        )
        warn(msg, gapsWarning)

    for cmd, cmd_status in status.items():
        try:
            command_index = cmd_status.get(StatusField.PIPELINE_INDEX)
        except AttributeError:
            continue

        if str(index) == str(command_index):
            return Status.parse_step_status(status_dir, cmd, key)

    return []



"""GAPs CLI template config generation command"""

import logging
import shutil
from pathlib import Path
from warnings import warn

import click

from rex.utilities.loggers import init_logger
from gaps.status import Status, StatusOption
from gaps.cli.command import _WrappedCommand
from gaps.warn import gapsWarning


logger = logging.getLogger(__name__)


@click.pass_context
def _reset_status(ctx, directory, force=False, after_step=None):
    """Filter configs and write to file based on type"""
    if ctx.obj.get("VERBOSE"):
        init_logger("gaps")

    if not directory:
        directory = [Path("./")]

    for status_dir in directory:
        status_dir = Path(status_dir).expanduser().resolve()  # noqa: PLW2901
        status_file_dir = status_dir / Status.HIDDEN_SUB_DIR
        if not status_file_dir.exists():
            logger.debug(
                "No status info detected in %r. Skipping...", str(status_dir)
            )
            continue

        status = Status(status_dir)
        is_processing = (
            status.as_df()
            .job_status.isin({StatusOption.SUBMITTED, StatusOption.RUNNING})
            .any()
        )
        if is_processing and not force:
            msg = (
                f"Found queued/running jobs in {status_dir}. "
                "Not resetting... (override this behavior with --force)"
            )
            warn(msg, gapsWarning)
            continue

        if after_step:
            if after_step not in status.data:
                msg = (
                    f"Command {after_step!r} not found as part of pipeline "
                    f"in {status_dir}. Not resetting..."
                )
                warn(msg, gapsWarning)
                continue
            logger.info("Resetting status for all steps after %r", after_step)
            status.update_from_all_job_files()
            status.reset_after(after_step)
            status.dump()
        else:
            logger.info(
                "Removing status info for directory %r", str(status_dir)
            )
            shutil.rmtree(status_file_dir)


def reset_command():
    """A status reset CLI command"""
    params = [
        click.Argument(
            param_decls=["directory"],
            required=False,
            nargs=-1,
        ),
        click.Option(
            param_decls=["--force", "-f"],
            is_flag=True,
            help="Force pipeline status reset even if jobs are queued/running",
        ),
        click.Option(
            param_decls=["--after-step", "-a"],
            multiple=False,
            default=None,
            help="Reset pipeline starting after the given pipeline step. The "
            "status of this step will remain unaffected, but the status of "
            "steps following it will be reset completely.",
        ),
    ]
    return _WrappedCommand(
        "reset-status",
        context_settings=None,
        callback=_reset_status,
        params=params,
        help=(
            "Reset the pipeline/job status (progress) for a given directory "
            "(defaults to ``./``). Multiple directories can be supplied to "
            "reset the status of each. \n\nThe general structure for calling "
            "this CLI command is given below (add ``--help`` to print help "
            "info to the terminal)."
        ),
        epilog=None,
        short_help=None,
        options_metavar="",
        add_help_option=True,
        no_args_is_help=False,
        hidden=False,
        deprecated=False,
    )



"""GAPs execution CLI utils"""

import logging
import datetime as dt
from pathlib import Path
from copy import deepcopy
from warnings import warn
from inspect import signature

from gaps.hpc import submit, DEFAULT_STDOUT_PATH
from gaps.status import (
    DT_FMT,
    Status,
    HardwareOption,
    StatusOption,
    StatusField,
    QOSOption,
)
from gaps.warn import gapsWarning
from gaps.exceptions import gapsConfigError

logger = logging.getLogger(__name__)


def kickoff_job(ctx, cmd, exec_kwargs):
    """Kickoff a single job (a single command execution).

    Parameters
    ----------
    ctx : click.Context
        Context object with a `.obj` attribute that contains at least
        the following keys:

            NAME : str
                Job name.
            OUT_DIR : path-like
                Path to output directory.
            COMMAND_NAME : str
                Name of command being run.
            PIPELINE_STEP: str
                Name of pipeline step being run.

    cmd : str
        String form of command to kickoff.
    exec_kwargs : dict
        Keyword-value pairs to pass to the respective `submit` function.
        These will be filtered, so they may contain extra values. If
        some required inputs are missing from this dictionary, a
        `gapsConfigError` is raised.

    Raises
    ------
    gapsConfigError
        If `exec_kwargs` is missing some arguments required by the
        respective `submit` function.
    """
    exec_kwargs = deepcopy(exec_kwargs)
    hardware_option = HardwareOption(exec_kwargs.pop("option", "local"))
    if hardware_option.manager is None:
        _kickoff_local_job(ctx, cmd)
        return

    ctx.obj["MANAGER"] = hardware_option.manager
    exec_kwargs = _filter_exec_kwargs(
        exec_kwargs, hardware_option.manager.make_script_str, hardware_option
    )
    _kickoff_hpc_job(ctx, cmd, hardware_option, **exec_kwargs)


def _filter_exec_kwargs(kwargs, func, hardware_option):
    """Filter out extra keywords and raise error if any are missing"""
    sig = signature(func)
    kwargs_to_use = {k: v for k, v in kwargs.items() if k in sig.parameters}
    extra_keys = set(kwargs) - set(kwargs_to_use)
    if extra_keys:
        msg = (
            f"Found extra keys in 'execution_control'! The following "
            f"inputs will be ignored: {extra_keys}. To silence this warning, "
            "please remove the extra keys from the 'execution_control' block."
        )
        warn(msg, gapsWarning)

    required = {
        name for name, p in sig.parameters.items() if p.default == p.empty
    }
    required -= {"self", "cmd", "name"}
    missing = {k for k in required if k not in kwargs_to_use}
    if missing:
        msg = (
            f"The 'execution_control' block is missing the following "
            f"required keys: {missing}"
        )
        raise gapsConfigError(msg)

    if hardware_option.supports_categorical_qos:
        qos = kwargs_to_use.get("qos", "normal")
        try:
            qos = QOSOption(qos)
        except ValueError as err:
            msg = (
                f"Requested Quality-of-service option ({qos!r}) not "
                f"recognized! Available options are: "
                f"{QOSOption.members_as_str()}."
            )
            raise gapsConfigError(msg) from err

        kwargs_to_use["qos"] = f"{qos}"

    return kwargs_to_use


def _kickoff_local_job(ctx, cmd):
    """Run a job (command) locally"""

    if not _should_run(ctx):
        return

    name = ctx.obj["NAME"]
    command = ctx.obj["COMMAND_NAME"]
    logger.info("Running %r locally with job name %r.", command, name)
    logger.debug("Submitting the following command:\n%s", cmd)
    Status.mark_job_as_submitted(
        ctx.obj["OUT_DIR"],
        pipeline_step=ctx.obj["PIPELINE_STEP"],
        job_name=name,
        replace=True,
        job_attrs={
            StatusField.JOB_STATUS: StatusOption.SUBMITTED,
            StatusField.HARDWARE: HardwareOption.LOCAL,
            StatusField.TIME_SUBMITTED: dt.datetime.now().strftime(DT_FMT),
        },
    )
    stdout, stderr = submit(cmd)
    if stdout:
        logger.info("Subprocess received stdout: \n%s", stdout)
    if stderr:
        logger.warning("Subprocess received stderr: \n%s", stderr)
    msg = f"Completed job {name!r}."
    logger.info(msg)


def _kickoff_hpc_job(ctx, cmd, hardware_option, **kwargs):
    """Run a job (command) on the HPC"""

    if not _should_run(ctx):
        return

    name = ctx.obj["NAME"]
    command = ctx.obj["COMMAND_NAME"]
    logger.debug("Submitting the following command:\n%s", cmd)
    out = ctx.obj["MANAGER"].submit(name, cmd=cmd, **kwargs)[0]
    id_msg = f" (Job ID #{out})" if out else ""
    msg = f"Kicked off {command!r} job {name!r}{id_msg}"

    stdout_dir = Path(kwargs.get("stdout_path", DEFAULT_STDOUT_PATH))
    stdout_log_file = str(stdout_dir / f"{name}_{out}.o")
    stdout_err_log_file = str(stdout_dir / f"{name}_{out}.e")

    Status.mark_job_as_submitted(
        ctx.obj["OUT_DIR"],
        pipeline_step=ctx.obj["PIPELINE_STEP"],
        job_name=name,
        replace=True,
        job_attrs={
            StatusField.JOB_ID: out,
            StatusField.HARDWARE: hardware_option,
            StatusField.QOS: kwargs.get("qos") or QOSOption.UNSPECIFIED,
            StatusField.JOB_STATUS: StatusOption.SUBMITTED,
            StatusField.TIME_SUBMITTED: dt.datetime.now().strftime(DT_FMT),
            StatusField.STDOUT_LOG: stdout_log_file,
            StatusField.STDOUT_ERR_LOG: stdout_err_log_file,
        },
    )
    logger.info(msg)


def _should_run(ctx):
    """Determine whether a command should be run based on status"""
    name = ctx.obj["NAME"]
    out_dir = ctx.obj["OUT_DIR"]
    status = Status.retrieve_job_status(
        out_dir,
        pipeline_step=ctx.obj["PIPELINE_STEP"],
        job_name=name,
        subprocess_manager=ctx.obj.get("MANAGER"),
    )
    if status == StatusOption.NOT_SUBMITTED:
        return True

    if status in {StatusOption.SUCCESSFUL, StatusOption.COMPLETE}:
        msg = (
            f"Job {name!r} is successful in status json found in {out_dir!r}, "
            f"not re-running."
        )
        logger.info(msg)
        return False

    if status is not None and "fail" not in str(status).lower():
        msg = (
            f"Job {name!r} was found with status {status!r}, not resubmitting"
        )
        logger.info(msg)
        return False

    return True



"""Batch Job CLI entry points"""

import click

import gaps.batch
from gaps.log import init_logger
from gaps.config import init_logging_from_config_file
from gaps.cli.command import _WrappedCommand
from gaps.cli.documentation import _batch_command_help


def _batch(config_file, dry, cancel, delete, monitor_background):
    """Execute an analysis pipeline over a parametric set of inputs"""
    if str(config_file).endswith("csv"):
        init_logger(stream=not monitor_background, level="INFO", file=None)
    else:
        init_logging_from_config_file(
            config_file, background=monitor_background
        )

    if cancel:
        gaps.batch.BatchJob(config_file).cancel()
    elif delete:
        gaps.batch.BatchJob(config_file).delete()
    else:
        gaps.batch.BatchJob(config_file).run(
            dry_run=dry,
            monitor_background=monitor_background,
        )


def batch_command():
    """Generate a batch command"""
    params = [
        click.Option(
            param_decls=["--config_file", "-c"],
            required=True,
            type=click.Path(exists=True),
            help=_batch_command_help(),
        ),
        click.Option(
            param_decls=["--dry"],
            is_flag=True,
            help="Flag to do a dry run (make batch dirs and update files "
            "without running the pipeline).",
        ),
        click.Option(
            param_decls=["--cancel"],
            is_flag=True,
            help="Flag to cancel all jobs associated associated with the "
            "``batch_jobs.csv`` file in the current batch config directory.",
        ),
        click.Option(
            param_decls=["--delete"],
            is_flag=True,
            help="Flag to delete all batch job sub directories associated "
            "with the ``batch_jobs.csv`` file in the current batch config "
            "directory.",
        ),
        click.Option(
            param_decls=["--monitor-background"],
            is_flag=True,
            help="Flag to monitor all batch pipelines continuously in the "
            "background. Note that the ``stdout/stderr`` will not be "
            'captured, but you can set a pipeline ``"log_file"`` to capture '
            "logs.",
        ),
    ]
    return _WrappedCommand(
        "batch",
        context_settings=None,
        callback=_batch,
        params=params,
        help=(
            "Execute an analysis pipeline over a parametric set of inputs.\n\n"
            "The general structure for calling this CLI command is given "
            "below (add ``--help`` to print help info to the terminal)."
        ),
        epilog=None,
        short_help=None,
        options_metavar="[OPTIONS]",
        add_help_option=True,
        no_args_is_help=True,
        hidden=False,
        deprecated=False,
    )
